# coding: utf-8

# flake8: noqa

# import all models into this package
# if you have many models here with many references from one model to another this may
# raise a RecursionError
# to avoid this, import only the models that you directly need like:
# from from domino._impl.custommetrics.model.pet import Pet
# or import this package, but before doing it, use:
# import sys
# sys.setrecursionlimit(n)

from domino._impl.custommetrics.model.failure_envelope_v1 import FailureEnvelopeV1
from domino._impl.custommetrics.model.invalid_body_envelope_v1 import InvalidBodyEnvelopeV1
from domino._impl.custommetrics.model.metadata_v1 import MetadataV1
from domino._impl.custommetrics.model.metric_alert_request_v1 import MetricAlertRequestV1
from domino._impl.custommetrics.model.metric_tag_v1 import MetricTagV1
from domino._impl.custommetrics.model.metric_value_v1 import MetricValueV1
from domino._impl.custommetrics.model.metric_values_envelope_v1 import MetricValuesEnvelopeV1
from domino._impl.custommetrics.model.new_metric_value_v1 import NewMetricValueV1
from domino._impl.custommetrics.model.new_metric_values_envelope_v1 import NewMetricValuesEnvelopeV1
from domino._impl.custommetrics.model.target_range_v1 import TargetRangeV1



import uuid
from pprint import pformat

import gzip
import pytest

from domino import Domino, exceptions
from domino.exceptions import ProjectNotFoundException
from domino.helpers import domino_is_reachable


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_project_create(default_domino_client):
    """
    Confirm that the python-domino client can create a new project.
    """
    new_project_name = f"project-{str(uuid.uuid4())}"
    response = default_domino_client.project_create(new_project_name)
    assert response.status_code == 200, f"{response.status_code}: {response.reason}"

    project_list = default_domino_client.projects_list()
    assert any(
        p["name"] == new_project_name for p in project_list
    ), f"Unable to retrieve new project!\n{pformat(project_list)}"

    default_domino_client.project_archive(new_project_name)


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_project_fork(default_domino_client):
    """
    Confirm that the python-domino client can fork an existing project.
    """
    forked_project_name = f"forked-project-{str(uuid.uuid4())}"
    response = default_domino_client.fork_project(forked_project_name)
    assert response.status_code == 200, f"{response.status_code}: {response.reason}"

    project_list = default_domino_client.projects_list()
    assert any(
        p["name"] == forked_project_name for p in project_list
    ), f"Unable to retrieve forked project!\n{pformat(project_list)}"

    default_domino_client.project_archive(forked_project_name)


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_list_commits(default_domino_client):
    """
    Confirm that the python-domino client can list project commits.
    """
    commits_list = default_domino_client.commits_list()
    assert isinstance(
        commits_list, list
    ), f"Unable to retrieve commits:\n{pformat(commits_list)}"


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_list_files_in_commit(default_domino_client):
    """
    Confirm that the python-domino client can list the files of a given project commit.
    """
    commits_list = default_domino_client.commits_list()
    files_list = default_domino_client.files_list(commits_list[0])
    assert isinstance(
        files_list["data"], list
    ), f"Unable to retrieve files:\n{pformat(files_list)}"


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_upload_file_to_project(default_domino_client):
    """
    Confirm that the python-domino client can upload a file to a project.
    """
    with open(__file__, "rb") as test_file:
        response = default_domino_client.files_upload(
            path="/test_file.py", file=test_file
        )
    assert response.status_code == 201
    assert response.json()["path"] == "test_file.py"


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_upload_file_to_project_without_forward_slash(default_domino_client):
    """
    Confirm that the python-domino client can upload a file to a project.
    """
    with open(__file__, "rb") as test_file:
        response = default_domino_client.files_upload(
            path="test_file.py", file=test_file
        )
    assert response.status_code == 201
    assert response.json()["path"] == "test_file.py"


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_get_file_from_a_project(default_domino_client):
    """
    Confirm that the python-domino client can download a file from a project.
    """
    commits_list = default_domino_client.commits_list()
    files_list = default_domino_client.files_list(commits_list[0])

    for file in files_list["data"]:
        if file["path"] == ".dominoignore":
            file_contents = default_domino_client.blobs_get(file["key"]).read()
            break

    assert "ignore certain files" in str(
        file_contents
    ), f"Unable to get .dominoignore file\n{str(file_contents)}"


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_get_file_from_a_project_v2(default_domino_client):
    """
    Confirm that the python-domino client can download a file from a project in the v2 endpoint
    """
    commits_list = default_domino_client.commits_list()
    files_list = default_domino_client.files_list(commits_list[0])

    for file in files_list["data"]:
        if file["path"] == ".dominoignore":
            file_contents = default_domino_client.blobs_get_v2(file["path"], commits_list[0], default_domino_client.project_id).read()
            break

    assert "ignore certain files" in str(
        file_contents
    ), f"Unable to get .dominoignore file\n{str(file_contents)}"


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_get_blobs_v2_non_canonical(default_domino_client):
    """
    Confirm that the python-domino client get_blobs_v2 will fail if input path is non-canonical
    """
    non_canonical_path = "/domino/mnt/../test.py"
    commits_list = default_domino_client.commits_list()

    with pytest.raises(exceptions.MalformedInputException):
        default_domino_client.blobs_get_v2(non_canonical_path, commits_list[0], default_domino_client.project_id).read()


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_add_and_remove_project_collaborator(default_domino_client):
    """
    Confirm that the python-domino client can and/remove project collaborators.
    """

    def _get_users():
        # This operation is not yet a public-facing part of the python-domino API
        url = default_domino_client._routes.users_get()
        response = default_domino_client.request_manager.get(url)
        assert response.status_code == 200, f"{response.status_code}: {response.reason}"
        return response.json()

    users_list = _get_users()
    collaborators = default_domino_client.collaborators_get()

    # Find a username that is not already a collaborator
    for user in users_list:
        username = user["userName"]
        if username not in collaborators:
            default_domino_client.collaborators_add(username)
            break

    new_collaborators = default_domino_client.collaborators_get()
    assert username in new_collaborators


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_publish_app_from_a_project(default_domino_client):
    """
    Confirm that the python-domino client can publish an app from a project.
    """
    response = default_domino_client.app_publish()
    assert response.status_code == 200, f"{response.status_code}: {response.reason}"


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_unpublish_app_from_a_project(default_domino_client):
    """
    Confirm that the python-domino client can unpublish an app from a project.
    """
    response = default_domino_client.app_unpublish()
    assert response.status_code == 200, f"{response.status_code}: {response.reason}"


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_tags_list_add_to_project(default_domino_client):
    """
    Confirm that the python-domino client can get and add tags a project.
    """
    first_tags = default_domino_client.tags_list()

    default_domino_client.tags_add(["new-tags"])
    new_tags = default_domino_client.tags_list()

    assert len(new_tags) > len(first_tags)

    assert any("new-tags" == tag["name"] for tag in new_tags)


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_tags_details(default_domino_client):
    """
    Confirm that the python-domino client can get tags details.
    """
    tag_name = "test-detail-tags"
    test_tag = default_domino_client.tags_add([tag_name])
    tag_id = test_tag.json()[0]["id"]
    detail_tag = default_domino_client.tag_details(tag_id)

    assert detail_tag["name"] == tag_name
    assert detail_tag["_id"] == tag_id
    assert detail_tag["lastAdded"] is not None

    default_domino_client.tags_remove(tag_name)


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_tags_list_remove_from_a_project(default_domino_client):
    """
    Confirm that the python-domino client can get and add tags a project.
    """
    first_tags = default_domino_client.tags_list()

    default_domino_client.tags_remove("new-tags")
    new_tags = default_domino_client.tags_list()

    assert len(new_tags) < len(first_tags)

    assert ("new-tags" != tag["id"] for tag in new_tags)


def test_archiving_non_existent_project_raises_appropriate_error(
    dummy_hostname, requests_mock
):
    """
    Confirm that trying to archive a bogus project will throw the appropriate exception.
    """
    requests_mock.get(f"{dummy_hostname}/version", json={"version": "9.9.9"})
    d = Domino(host=dummy_hostname, project="anyuser/anyproject", api_key="whatever")

    projects_list_endpoint = "v4/gateway/projects?relationship=Owned&showCompleted=true"

    with pytest.raises(ProjectNotFoundException):
        requests_mock.get(
            f"{dummy_hostname}/{projects_list_endpoint}", json=[], status_code=200
        )
        d.project_archive("bogus_project")



import typing_extensions

from domino._impl.custommetrics.paths import PathValues
from domino._impl.custommetrics.apis.paths.api_metric_alerts_v1 import ApiMetricAlertsV1
from domino._impl.custommetrics.apis.paths.api_metric_values_v1 import ApiMetricValuesV1
from domino._impl.custommetrics.apis.paths.api_metric_values_v1_model_monitoring_id_metric import ApiMetricValuesV1ModelMonitoringIdMetric

PathToApi = typing_extensions.TypedDict(
    'PathToApi',
    {
        PathValues.API_METRIC_ALERTS_V1: ApiMetricAlertsV1,
        PathValues.API_METRIC_VALUES_V1: ApiMetricValuesV1,
        PathValues.API_METRIC_VALUES_V1_MODEL_MONITORING_ID_METRIC: ApiMetricValuesV1ModelMonitoringIdMetric,
    }
)

path_to_api = PathToApi(
    {
        PathValues.API_METRIC_ALERTS_V1: ApiMetricAlertsV1,
        PathValues.API_METRIC_VALUES_V1: ApiMetricValuesV1,
        PathValues.API_METRIC_VALUES_V1_MODEL_MONITORING_ID_METRIC: ApiMetricValuesV1ModelMonitoringIdMetric,
    }
)



import os

import pytest


@pytest.mark.skipif(
    os.getenv("DATA_SDK") == "yes", reason="Testing failure when dependency is missing"
)
def test_data_sources_import_fails():  # noqa
    try:
        from domino.data_sources import DataSourceClient

        assert DataSourceClient
    except ImportError as exc:
        assert "Please pip install dominodatalab-data" in str(exc)
    else:
        assert (
            False
        ), "Import should be failing when dominodatalab-data is not installed"



from datetime import datetime, timedelta
import inspect
import logging as logger
import os
import pytest
import threading
import time
from unittest.mock import call, patch

from ...conftest import TEST_AI_SYSTEMS_ENV_VARS
from domino.aisystems._constants import EXPERIMENT_AI_SYSTEM_TAG
from .mlflow_fixtures import fixture_create_prod_traces, create_span_at_time
from .test_util import reset_prod_tracing
from domino.aisystems._client import client
from domino.aisystems.tracing._util import build_ai_system_experiment_name
# NOTE: don't use this import to test public functions, use the tracing pytest fixture instead
from domino.aisystems.tracing.tracing import _search_traces
from domino.aisystems._eval_tags import InvalidEvaluationLabelException

def test_init_tracing_prod(setup_mlflow_tracking_server, mocker, mlflow, tracing):
        """
        should initialize autologging only once for each framework
        should create an experiment for the ai system and tag it only once
        """
        app_id = "appid"
        test_case_vars = {"DOMINO_AI_SYSTEM_IS_PROD": "true", "DOMINO_APP_ID": app_id}
        expected_experiment_name = build_ai_system_experiment_name(app_id)
        env_vars = TEST_AI_SYSTEMS_ENV_VARS | test_case_vars

        import domino.aisystems.tracing.tracing
        import domino.aisystems._client
        import mlflow
        autolog_spy = mocker.spy(domino.aisystems.tracing.inittracing, "call_autolog")
        set_experiment_tag_spy = mocker.spy(domino.aisystems._client.client, "set_experiment_tag")
        set_experiment_spy = mocker.spy(mlflow, "set_experiment")

        reset_prod_tracing()

        with patch.dict(os.environ, env_vars, clear=True):
                tracing.init_tracing(["sklearn"])
                tracing.init_tracing(["sklearn"])
                found_exp = mlflow.get_experiment_by_name(expected_experiment_name)

                assert autolog_spy.call_args_list == [call('sklearn')]
                assert set_experiment_tag_spy.call_count == 1, "should only save tag on experiment once"
                assert set_experiment_spy.call_count is not 0, "should set an active experiment"
                assert found_exp is not None, "ai system experiment should exist"
                assert found_exp.tags.get(EXPERIMENT_AI_SYSTEM_TAG) == "true", "ai system experiment should be tagged"

def test_init_tracing_logs_experiment_creation_debug(setup_mlflow_tracking_server, mlflow, tracing, caplog):
        """
        when log level is debug, verify the experiment creation log includes the experiment ID
        """
        app_id = "app_id_logs_debug"
        test_case_vars = {"DOMINO_AI_SYSTEM_IS_PROD": "true", "DOMINO_APP_ID": app_id}
        env_vars = TEST_AI_SYSTEMS_ENV_VARS | test_case_vars

        reset_prod_tracing()

        with patch.dict(os.environ, env_vars, clear=True), caplog.at_level(logger.DEBUG):
                tracing.init_tracing()
                expected_experiment_name = build_ai_system_experiment_name(app_id)
                exp = mlflow.get_experiment_by_name(expected_experiment_name)
                assert exp is not None, "experiment should be created in prod mode"
                assert f"Created experiment for AI System with ID {exp.experiment_id}" in caplog.text

def test_logging_traces_prod(setup_mlflow_tracking_server, mocker, mlflow, tracing):
        """
        traces created in separate threads forked from the same main thread
        should be saved to the same ai system experiment
        """
        app_id = "threaded_app_id"
        test_case_vars = {"DOMINO_AI_SYSTEM_IS_PROD": "true", "DOMINO_APP_ID": app_id}
        expected_experiment_name = build_ai_system_experiment_name(app_id)
        env_vars = TEST_AI_SYSTEMS_ENV_VARS | test_case_vars

        reset_prod_tracing()

        with patch.dict(os.environ, env_vars, clear=True):
                tracing.init_tracing()

                @tracing.add_tracing(name="a")
                def a(num):
                        return num

                @tracing.add_tracing(name="b")
                def b(num):
                        return num

                t1 = threading.Thread(target=a, args=(10,))
                t2 = threading.Thread(target=b, args=(10,))

                t1.start()
                t2.start()

                t1.join()
                t2.join()

        # a and b traces should all be in the ai system experiment
        traces_a = mlflow.search_traces(filter_string="trace.name = 'a'", return_type='list')
        traces_b = mlflow.search_traces(filter_string="trace.name = 'b'", return_type='list')

        def get_experiment_id(trace):
                return trace.info.trace_location.mlflow_experiment.experiment_id

        found_exp_ids = set([get_experiment_id(t) for t in traces_a + traces_b])
        actual_exp_id = set([mlflow.get_experiment_by_name(expected_experiment_name).experiment_id])
        assert found_exp_ids == actual_exp_id, "traces should be linked to the ai system experiment"


def test_init_tracing_dev_mode(setup_mlflow_tracking_server, mocker, mlflow, tracing):
        """
        should not create an experiment or set tags
        """
        import domino.aisystems._client
        import mlflow
        set_experiment_tag_spy = mocker.spy(domino.aisystems._client.client, "set_experiment_tag")
        set_experiment_spy = mocker.spy(mlflow, "set_experiment")

        with patch.dict(os.environ, TEST_AI_SYSTEMS_ENV_VARS, clear=True):
                tracing.init_tracing(["sklearn"])

                assert set_experiment_tag_spy.call_count == 0, "should set experiment tag"
                assert set_experiment_spy.call_count == 0, "should not set an active experiment"

def test_add_tracing_dev(setup_mlflow_tracking_server, mocker, mlflow, tracing, logging):
        """
        add_tracing will create a new trace with a given name
        and attach evaluation tags to the trace
        """
        # must import logging from the module instead of the package
        # so that mocker works
        exp = mlflow.set_experiment("test_add_tracing_dev")

        @tracing.add_tracing(name="add_numbers", autolog_frameworks=["sklearn"], evaluator=lambda span: { 'result': span.outputs })
        def add_numbers(x, y):
                return x + y

        with logging.DominoRun("test_add_tracing_dev"):
                add_numbers(1, 1)

        ts = mlflow.search_traces(experiment_ids=[exp.experiment_id], return_type='list')
        assert len(ts) == 1, "only one trace should be created"

        # assert tags
        tags = ts[0].info.tags
        assert tags['domino.prog.metric.result'] == '2'
        assert tags['domino.internal.is_eval'] == 'true'

def test_add_tracing_dev_use_trace_in_evaluator(setup_mlflow_tracking_server, mocker, mlflow, tracing, logging, caplog):
        """
        User can access a trace in an inline evaluator on the trace parent, and not the child
        """
        exp = mlflow.set_experiment("test_add_tracing_dev_use_trace_in_evaluator")

        @tracing.add_tracing(name="parent", evaluator=lambda span: { 'span_exists_parent': 'True' }, trace_evaluator=lambda trace: { 'trace_exists_parent': 'True' })
        def parent(x):
                return unit(x)

        def child_trace_evaluator(trace):
                return { 'trace_exists_child': 'True' }

        @tracing.add_tracing(name="unit", evaluator=lambda span: { 'span_exists_child': 'True' }, trace_evaluator=child_trace_evaluator)
        def unit(x):
                return x

        with logging.DominoRun() as run, caplog.at_level(logger.WARNING):
                parent(1)

        parent_t = tracing.search_traces(run_id=run.info.run_id, trace_name="parent").data[0]

        evals = {r.name: r.value for r in parent_t.evaluation_results}
        assert evals.get('trace_exists_parent') == 'True'
        assert 'trace_exists_child' not in evals
        assert evals.get('span_exists_parent') == 'True'
        assert evals.get('span_exists_child') == 'True'
        assert "A trace_evaluator child_trace_evaluator was provided, but the trace could not be found" in caplog.text

def test_add_tracing_invalid_label(setup_mlflow_tracking_server, tracing):
        with pytest.raises(InvalidEvaluationLabelException):
                @tracing.add_tracing(name="*")
                def unit(x):
                        return x

def test_add_tracing_dev_no_evaluator(setup_mlflow_tracking_server, mlflow, tracing, logging):
        """
        add_tracing will create a new trace not add evaluations
        """
        exp = mlflow.set_experiment("test_add_tracing_dev_no_evaluator")

        @tracing.add_tracing(name="add_numbers")
        def add_numbers(x, y):
                return x + y

        with logging.DominoRun("test_add_tracing_dev_no_evaluator"):
                add_numbers(1, 1)

        # assert tags
        ts = mlflow.search_traces(experiment_ids=[exp.experiment_id], return_type='list')
        tags = ts[0].info.tags

        assert 'domino.internal.is_eval' not in tags

def test_add_tracing_decorator_preserves_function_info(setup_mlflow_tracking_server, tracing):
        def func_with_args(a: int, b: int, c: int=10, *args, **kwargs):
                """Function with various parameter types."""
                return a + b + c

        @tracing.add_tracing(name="decorated_func")
        def decorated_func(a: int, b: int, c: int=10, *args, **kwargs):
                """returns the input value"""
                return a + b + c

        original_sig = inspect.signature(func_with_args)
        decorated_sig = inspect.signature(decorated_func)

        assert decorated_func.__name__ == "decorated_func", "the function name should be preserved by the decorator"
        assert decorated_func.__doc__ == "returns the input value", "the function docstring should be preserved by the decorator"
        assert decorated_func.__module__ == "tests.integration.aisystems.test_aisystems_tracing"
        assert decorated_sig == original_sig
        assert list(decorated_sig.parameters.keys()) == ['a', 'b', 'c', 'args', 'kwargs']
        assert decorated_sig.parameters['c'].default == 10
        assert decorated_sig.parameters['a'].annotation == int

def test_add_tracing_preseves_self_and_cls(setup_mlflow_tracking_server, tracing):
        """
        add_tracing should preserve self and cls for functionality of the decorated method
        """
        class MyClass:
                class_value = 2
                def __init__(self):
                        self.value = 1

                @tracing.add_tracing(name="instance_method")
                def instance_method(self, x):
                        return self.value + x

                @classmethod
                @tracing.add_tracing(name="class_method")
                def class_method(cls, x):
                        return cls.class_value + x

        obj = MyClass()

        assert obj.instance_method(1) == 2
        assert MyClass.class_method(1) == 3

def test_add_tracing_arguments_passed_to_span(setup_mlflow_tracking_server, tracing, mlflow):
        """
        add_tracing should preserve self and cls for functionality of the decorated method,
        but should not pass them as inputs to the trace.

        it should pass args and kwargs as inputs, also
        it should pass default values
        """
        exp = mlflow.set_experiment("test_add_tracing_arguments_passed_to_span")
        experiment_id = exp.experiment_id

        class MyClass:
                @tracing.add_tracing(name="instance_method")
                def instance_method(self, x):
                        return x

                @classmethod
                @tracing.add_tracing(name="class_method")
                def class_method(cls, x):
                        return x

        @tracing.add_tracing(name="args_kwargs")
        def args_kwargs(*args, **kwargs):
                return (args, kwargs)

        @tracing.add_tracing(name="fun_with_defaults")
        def fun_with_defaults(x=10):
                return x

        obj = MyClass()
        obj.instance_method(1)
        MyClass.class_method(1)
        args_kwargs(1, y=2)
        fun_with_defaults()


        instance_trace = mlflow.search_traces(experiment_ids=[experiment_id], return_type='list', filter_string="trace.name = 'instance_method'")[0]
        class_trace = mlflow.search_traces(experiment_ids=[experiment_id], return_type='list', filter_string="trace.name = 'class_method'")[0]
        args_kwargs_trace = mlflow.search_traces(experiment_ids=[experiment_id], return_type='list', filter_string="trace.name = 'args_kwargs'")[0]
        fun_with_defaults_trace = mlflow.search_traces(experiment_ids=[experiment_id], return_type='list', filter_string="trace.name = 'fun_with_defaults'")[0]

        def get_inputs(trace):
                return trace.data.spans[0].inputs

        it_inputs = get_inputs(instance_trace)
        assert it_inputs == {'x': 1}

        ct_inputs = get_inputs(class_trace)
        assert ct_inputs == {'x': 1}

        ak_inputs = get_inputs(args_kwargs_trace)
        assert ak_inputs == {'args': [1], 'kwargs': {'y': 2}}

        d_inputs = get_inputs(fun_with_defaults_trace)
        assert d_inputs == {'x': 10}

def test_add_tracing_failed_inline_evaluator_logs_warning(setup_mlflow_tracking_server, tracing, mlflow, caplog):
        """
        if the inline evaluator fails, a warning is logged and the main code still executes
        """
        mlflow.set_experiment("test_add_tracing_failed_inline_evaluator_logs_warning")

        def failing_trace_evaluator(t):
                return 1/0

        def failing_evaluator(span):
                return 1/0

        @tracing.add_tracing(name="unit", evaluator=failing_evaluator, trace_evaluator=failing_trace_evaluator)
        def unit(x):
                return x

        with mlflow.start_run(), caplog.at_level(logger.ERROR):
                assert unit(1) == 1
                print(caplog.text)
                assert "Inline evaluation failed for evaluator, failing_evaluator" in caplog.text
                assert "Inline evaluation failed for trace_evaluator, failing_trace_evaluator" in caplog.text

def test_add_tracing_works_with_generator(setup_mlflow_tracking_server, tracing, mlflow):
        """
        add_tracing should not record all result from a generator if not specified
        if we don't eagerly load the reults onto one trace, we save a span for each yield
        """
        exp = mlflow.set_experiment("test_add_tracing_works_with_generator")
        experiment_id = exp.experiment_id

        @tracing.add_tracing(name="gen", evaluator=lambda span: { 'result': 1 }, eagerly_evaluate_streamed_results=False)
        def gen():
                for i in range(3):
                        yield i

        xs = [x for x in gen()]
        assert xs == [0, 1, 2], "Results should be unaffected by tracing"

        gen_trace = mlflow.search_traces(experiment_ids=[experiment_id], return_type='list', filter_string="trace.name = 'gen'")[0]
        assert len(gen_trace.data.spans) == 4, "should have 4 spans, one for function call, and one for each yield"
        assert [s.outputs for s in gen_trace.data.spans[1:]] == [0, 1, 2], "yields spans should have correct outputs"
        assert ["group_id" in s.attributes for s in gen_trace.data.spans[1:]] == [True, True, True], "yields spans should have a group_id attribute"
        assert [s.attributes["index"] for s in gen_trace.data.spans[1:]] == [0, 1, 2]
        assert len(set([s.attributes["group_id"] for s in gen_trace.data.spans[1:]])) == 1, "group_id should be the same for all yields"

        # assert evaluation didn't happen inline
        tags = gen_trace.info.tags
        assert 'domino.prog.metric.result' not in tags
        assert 'domino.internal.is_eval' not in tags

def test_add_tracing_generator_trace_in_evaluator(setup_mlflow_tracking_server, tracing, mlflow, logging):
        """
        When using a generator, the trace should be accessible in the parent generator function's evaluator,
        but not the child span's evaluator
        """
        exp = mlflow.set_experiment("test_add_tracing_generator_trace_in_evaluator")
        experiment_id = exp.experiment_id

        @tracing.add_tracing(name="parent", evaluator=lambda span: { 'span_exists_parent': 'True' }, trace_evaluator=lambda trace: { 'trace_exists_parent': 'True' })
        def parent():
                yield from child(1)

        @tracing.add_tracing(name="child", evaluator=lambda span: { 'span_exists_child': 'True' }, trace_evaluator=lambda trace: { 'trace_exists_child': 'True' })
        def child(x):
                yield x


        with logging.DominoRun() as run:
                [_ for _ in parent()]

        parent_t = tracing.search_traces(run_id=run.info.run_id, trace_name="parent").data[0]
        evals = {r.name: r.value for r in parent_t.evaluation_results}
        assert evals.get('trace_exists_parent') == 'True'
        assert 'trace_exists_child' not in evals
        assert evals.get('span_exists_parent') == 'True'
        assert evals.get('span_exists_child') == 'True'

def test_add_tracing_works_with_eagerly_evaluated_generator(setup_mlflow_tracking_server, tracing, mlflow):
        """
        add_tracing should record the result from a generator and evaluate it inline
        """
        exp = mlflow.set_experiment("test_add_tracing_works_with_eagerly_evaluated_generator")
        experiment_id = exp.experiment_id

        @tracing.add_tracing(name="gen_record_all", evaluator=lambda span: { 'result': 1 })
        def gen_record_all():
                for i in range(3):
                        yield i

        xs = [x for x in gen_record_all()]
        assert xs == [0, 1, 2]

        gen_trace = mlflow.search_traces(experiment_ids=[experiment_id], return_type='list', filter_string="trace.name = 'gen_record_all'")[0]
        span = gen_trace.data.spans[0]
        tags = gen_trace.info.tags

        assert len(gen_trace.data.spans) == 1
        assert span.outputs == [0, 1, 2]
        assert tags['domino.prog.metric.result'] == '1'
        assert tags['domino.internal.is_eval'] == 'true'

@pytest.mark.asyncio
async def test_add_tracing_works_with_async(setup_mlflow_tracking_server, mlflow, tracing):
        exp = mlflow.set_experiment("test_add_tracing_works_with_async")

        @tracing.add_tracing(name="async_function", evaluator=lambda span: { 'result': 1 })
        async def async_function(x):
                return x

        res = await async_function(1)
        assert res == 1

        traces = mlflow.search_traces(experiment_ids=[exp.experiment_id], return_type='list')

        assert [t.data.spans[0].inputs for t in traces] == [{'x':1}], "Inputs to trace should be the function arguments"
        assert [t.data.spans[0].outputs for t in traces] == [1], "Outputs to trace should be the function return value"

@pytest.mark.asyncio
async def test_add_tracing_async_trace_in_evaluator(setup_mlflow_tracking_server, mlflow, tracing, logging):
        """
        When using async functions, the trace should be accessible in the parent function's evaluator
        but not the child function's evaluator
        """
        exp = mlflow.set_experiment("test_add_tracing_async_trace_in_evaluator")

        @tracing.add_tracing(name="parent", evaluator=lambda span: { 'span_exists_parent': 'True' }, trace_evaluator=lambda trace: { 'trace_exists_parent': 'True' })
        async def parent(x):
                return await child(x)

        @tracing.add_tracing(name="child", evaluator=lambda span: { 'span_exists_child': 'True' }, trace_evaluator=lambda trace: { 'trace_exists_child': 'True'})
        async def child(x):
                return x

        with logging.DominoRun() as run:
                await parent(1)

        parent_t = tracing.search_traces(run_id=run.info.run_id, trace_name="parent").data[0]
        parent_t = tracing.search_traces(run_id=run.info.run_id, trace_name="parent").data[0]
        evals = {r.name: r.value for r in parent_t.evaluation_results}
        assert evals.get('trace_exists_parent') == 'True'
        assert 'trace_exists_child' not in evals
        assert evals.get('span_exists_parent') == 'True'
        assert evals.get('span_exists_child') == 'True'

def test_search_traces(setup_mlflow_tracking_server, mocker, mlflow, tracing, logging):
        @tracing.add_tracing(name="unit")
        def unit(x):
                return x

        @tracing.add_tracing(name="parent", evaluator=lambda span: {'mymetric': 1, 'mylabel': 'category'})
        def parent(x, y):
                return unit(x) + unit(y)

        @tracing.add_tracing(name="parent2")
        def parent2(x):
                return x

        mlflow.set_experiment("test_search_traces")
        run_id = None
        with logging.DominoRun() as run:
                run_id = run.info.run_id
                parent(1, 2)
                parent2(1)

        res = tracing.search_traces(run_id=run_id)
        span_data = [(s.name, s.inputs, s.outputs) for trace in res.data for s in trace.spans]

        assert sorted([trace.name for trace in res.data]) == sorted(["parent", "parent2"])
        assert sorted([(t.name, t.value) for trace in res.data for t in trace.evaluation_results if trace.name == "parent"]) \
                == sorted([("mylabel", "category"), ("mymetric", 1.0)])
        assert sorted(span_data) == sorted([("parent", {'x':1, 'y': 2}, 3), \
                ("parent2", {'x':1}, 1), ("unit_1", {'x':1}, 1), ("unit_2", {'x':2}, 2)
        ])

def test_search_traces_time_filter_warning(setup_mlflow_tracking_server, tracing, mlflow, logging, caplog):
        """
        if start time is > end time, warn the user
        """
        mlflow.set_experiment("test_search_traces_time_filter_warning")
        run_id = None
        with logging.DominoRun() as run:
                run_id = run.info.run_id

        with caplog.at_level(logger.WARNING):
                tracing.search_traces(run_id=run_id, start_time=datetime.now(), end_time=datetime.now() - timedelta(seconds=10))
                assert f"start_time must be before end_time" in caplog.text

def test_search_traces_by_trace_name(setup_mlflow_tracking_server, mocker, mlflow, tracing, logging):
        @tracing.add_tracing(name="unit")
        def unit(x):
                return x

        @tracing.add_tracing(name="parent")
        def parent(x, y):
                return unit(x) + unit(y)

        @tracing.add_tracing(name="parent2")
        def parent2(x):
                return x

        mlflow.set_experiment("test_search_traces_by_trace_name")
        run_id = None
        with logging.DominoRun() as run:
                run_id = run.info.run_id
                parent(1, 2)
                parent2(1)

        res = tracing.search_traces(run_id=run_id, trace_name="parent")
        span_data = [(s.name, s.inputs, s.outputs) for trace in res.data for s in trace.spans]

        assert [trace.name for trace in res.data] == ["parent"]
        assert sorted(span_data) == sorted([("parent", {'x':1, 'y': 2}, 3), ("unit_1", {'x':1}, 1), ("unit_2", {'x':2}, 2)])

def test_search_traces_by_timestamp(setup_mlflow_tracking_server, mocker, mlflow, tracing, logging):
        @tracing.add_tracing(name="parent")
        def parent(x):
                return x

        mlflow.set_experiment("test_search_traces_by_timestamp")
        run_id = None
        with logging.DominoRun() as run:
                run_id = run.info.run_id
                parent(1)

                time.sleep(2)

                parent(2)

                time.sleep(2)

                parent(3)

        start_time = datetime.now() - timedelta(seconds=4)
        end_time = datetime.now() - timedelta(seconds=2)

        res = tracing.search_traces(
                run_id=run_id,
                trace_name="parent",
                start_time=start_time,
                end_time=end_time
        )

        assert [trace.name for trace in res.data] == ["parent"]
        assert [[(s.name, s.inputs['x'], s.outputs) for s in trace.spans] for trace in res.data] == [[("parent", 2, 2)]]

def test_search_traces_with_traces_made_2hrs_ago(setup_mlflow_tracking_server, mocker, mlflow, tracing, logging):
        exp = mlflow.set_experiment("test_search_traces_with_traces_made_2hrs_ago")

        def parent(x):
                dt = datetime.now() - timedelta(hours=2)
                ns = int(dt.timestamp() * 1e9)
                span = mlflow.start_span_no_context(name="parent",  inputs=1, experiment_id=exp.experiment_id, start_time_ns=ns)
                span.end()
                return x

        run_id = None
        with logging.DominoRun() as run:
                run_id = run.info.run_id
                parent(1)

        res = tracing.search_traces(
                run_id=run_id,
                trace_name="parent",
        )

        assert [trace.name for trace in res.data] == ["parent"]

        # If i shorten the time filter, I get no results
        recent_res = tracing.search_traces(
                run_id=run_id,
                trace_name="parent",
                start_time=datetime.now() - timedelta(hours=1),
        )
        assert recent_res.data == []

def test_search_traces_multiple_runs_in_exp(setup_mlflow_tracking_server, mocker, mlflow, tracing, logging):
        exp = mlflow.set_experiment("test_search_traces_multiple_runs_in_exp")

        @tracing.add_tracing(name="unit1")
        def unit1(x):
                return x

        @tracing.add_tracing(name="unit2")
        def unit2(x):
                return x

        run_1_id = None
        with logging.DominoRun() as run:
                run_1_id = run.info.run_id
                unit1(1)

        with logging.DominoRun() as run:
                unit2(1)

        res = tracing.search_traces(run_id=run_1_id)

        assert [trace.name for trace in res.data] == ["unit1"]

def test_search_traces_ai_system(setup_mlflow_tracking_server_no_env_var_mock, mlflow, tracing):
        """
        Can filter by ai system id alone or id and version
        """
        app_id = "test_search_traces_ai_system_id"
        app_version_1 = "test_search_traces_ai_system_version_1"
        app_version_2 = "test_search_traces_ai_system_version_2"

        fixture_create_prod_traces(app_id, app_version_1, "one", tracing)
        fixture_create_prod_traces(app_id, app_version_2, "two", tracing)

        def get_trace_names(traces):
                return sorted([trace.name for trace in traces.data])

        all_traces = tracing.search_ai_system_traces(ai_system_id=app_id)
        assert get_trace_names(all_traces) == ["one", "two"], "Can get traces for all ai system versions"

        v1_traces = tracing.search_ai_system_traces(ai_system_id=app_id, ai_system_version=app_version_1)
        assert get_trace_names(v1_traces) == ["one"], "Can get traces for just ai system version 1"

        v2_traces = tracing.search_ai_system_traces(ai_system_id=app_id, ai_system_version=app_version_2)
        assert get_trace_names(v2_traces) == ["two"], "Can get traces for just ai system version 2"

def test_search_traces_ai_system_ai_system_id_required(setup_mlflow_tracking_server_no_env_var_mock):
        """
        ai system id is required if version supplied
        """

        with pytest.raises(Exception) as e_info:
                _search_traces(ai_system_version="fakeversion")

        assert "ai_system_id must also be provided" in str(e_info), "Should raise if version provided without id"

def test_search_traces_no_run_ai_system_ids_supplied(setup_mlflow_tracking_server_no_env_var_mock, tracing):
        """
        should throw if no run id, ai system version, or id supplied
        """

        with pytest.raises(Exception) as e_info:
                _search_traces()

        assert "Either run_id or ai_system_id and ai_system_version must be provided to search traces" in str(e_info), \
                "Should raise no ai system info or run info provided"

def test_search_traces_filters_should_work_together_dev(setup_mlflow_tracking_server, mocker, mlflow, tracing, logging):
        """
        When every filter is specified as well as pagination, the expected results should be returned
        The test creates multiple differently named traces over the course of a few hours in an experiment
        with multiple runs
        """
        exp = mlflow.set_experiment("test_search_traces_filters_should_work_together_dev")

        @tracing.add_tracing(name="unit1")
        def unit1(x):
                return x

        def create_span_at_time(name: str, inputs: int, hours_ago: int):
                dt = datetime.now() - timedelta(hours=hours_ago)
                ns = int(dt.timestamp() * 1e9)
                span = mlflow.start_span_no_context(name=name, inputs=inputs, experiment_id=exp.experiment_id, start_time_ns=ns)
                span.end()

        @tracing.add_tracing(name="sum1")
        def sum1(x, y):
                return x + y

        @tracing.add_tracing(name="unit2")
        def unit2(x):
                return x

        run_1_id = None
        with logging.DominoRun() as run:
                run_1_id = run.info.run_id

                create_span_at_time(name="sum1", inputs=1, hours_ago=5)

                # search_traces should return the following two spans
                create_span_at_time(name="sum1", inputs=2, hours_ago=3)
                create_span_at_time(name="sum1", inputs=3, hours_ago=2)

                unit1(1)

        with logging.DominoRun() as run:
                unit2(1)

        start_time = datetime.now() - timedelta(hours=4)
        end_time = datetime.now() - timedelta(hours=1)

        def get_traces(next_page_token):
                return tracing.search_traces(
                        run_id=run_1_id,
                        trace_name="sum1",
                        start_time=start_time,
                        end_time=end_time,
                        page_token=next_page_token,
                        max_results=1
                )

        def get_span_data(page):
                return [(trace.name, [s.inputs for s in trace.spans]) for trace in page.data]

        # should only return the first sum1 call in the run_1_id domino run
        page1 = get_traces(None)
        assert get_span_data(page1) == [("sum1", [2])], "Should return first call"

        # should only return the second sum1 call in the run_1_id domino run
        page2 = get_traces(page1.page_token)
        assert get_span_data(page2) == [("sum1", [3])], "Should return second call"


def test_search_traces_filters_should_work_together_prod(setup_mlflow_tracking_server_no_env_var_mock, mocker, mlflow, tracing, logging):
        """
        When searching by ai system ID and version and when every filter is specified as well as pagination,
        the expected results should be returned
        The test creates multiple differently named traces over the course of a few hours in an experiment
        with multiple runs
        """
        app_id = "test_search_traces_filters_should_work_together_prod"
        app_version_1 = f"{app_id}_1"
        app_version_2 = f"{app_id}_2"

        fixture_create_prod_traces(app_id, app_version_1, "sum1", tracing, hours_ago=5)

        # search_traces should return the following two spans
        fixture_create_prod_traces(app_id, app_version_1, "sum1", tracing, hours_ago=2)
        fixture_create_prod_traces(app_id, app_version_1, "sum1", tracing, hours_ago=3)

        fixture_create_prod_traces(app_id, app_version_1, "unit1", tracing)
        fixture_create_prod_traces(app_id, app_version_2, "unit2", tracing)

        start_time = datetime.now() - timedelta(hours=4)
        end_time = datetime.now() - timedelta(hours=1)

        def get_traces(next_page_token):
                return tracing.search_ai_system_traces(
                        ai_system_id=app_id,
                        ai_system_version=app_version_1,
                        trace_name="sum1",
                        start_time=start_time,
                        end_time=end_time,
                        page_token=next_page_token,
                        max_results=1
                )

        def get_span_data(page):
                return [(trace.name, [s.inputs for s in trace.spans]) for trace in page.data]

        # should only return the first sum1 call
        page1 = get_traces(None)
        assert get_span_data(page1) == [("sum1", [3])], "Should return first call"

        # should only return the second sum1 call
        page2 = get_traces(page1.page_token)
        assert get_span_data(page2) == [("sum1", [2])], "Should return second call"


def test_search_traces_pagination(setup_mlflow_tracking_server, mocker, mlflow, tracing, logging):
        """
        The api should provide a page token in if the total number of results is bigger than the max results
        and you can use that token to get the next page of results
        """
        @tracing.add_tracing(name="parent")
        def parent(x):
                return x

        mlflow.set_experiment("test_search_traces_by_timestamp")
        run_id = None
        with logging.DominoRun() as run:
                run_id = run.info.run_id
                parent(1)
                parent(2)

        res1 = tracing.search_traces(
                run_id=run_id,
                max_results=1,
        )

        assert [[(s.name, s.inputs['x'], s.outputs) for s in trace.spans] for trace in res1.data] == [[("parent", 1, 1)]]

        res2 = tracing.search_traces(
                run_id=run_id,
                max_results=1,
                page_token=res1.page_token
        )

        assert [[(s.name, s.inputs['x'], s.outputs) for s in trace.spans] for trace in res2.data] == [[("parent", 2, 2)]]

def test_search_traces_from_lazy_generator(setup_mlflow_tracking_server, mocker, mlflow, tracing, logging):
        @tracing.add_tracing(name="parent", eagerly_evaluate_streamed_results=False)
        def parent():
                for i in range(3):
                        yield i

        mlflow.set_experiment("test_search_traces_from_lazy_generator")
        run_id = None
        with logging.DominoRun() as run:
                run_id = run.info.run_id
                # traces don't emit unless you consume generator
                [x for x in parent()]

        traces = tracing.search_traces(
                run_id=run_id,
        )

        assert len(traces.data) == 1
        assert len(traces.data[0].spans) == 4


def test_init_tracing_triggers_one_get_experiment_by_name_calls_in_threads(setup_mlflow_tracking_server, mlflow, tracing):
        """
        init_tracing should call mlflow.set_experiment once
        when invoked concurrently from two threads and traces should go to the
        right experiment anyway
        """
        app_id = "concurrency_app"
        env_vars = TEST_AI_SYSTEMS_ENV_VARS | {"DOMINO_AI_SYSTEM_IS_PROD": "true", "DOMINO_APP_ID": app_id}
        expected_experiment_name = build_ai_system_experiment_name(app_id)

        reset_prod_tracing()

        with patch.dict(os.environ, env_vars, clear=True):

                def send_traces():
                        tracing.init_tracing()

                        @tracing.add_tracing(name="do")
                        def do():
                                return 1

                        do()

                # Spy on mlflow.set_experiment to ensure it is called once
                with patch.object(
                        mlflow,
                        "set_experiment",
                        wraps=mlflow.set_experiment,
                ) as spy_set_experiment:
                        t1 = threading.Thread(target=send_traces)
                        t2 = threading.Thread(target=send_traces)

                        t1.start()
                        t2.start()
                        t1.join()
                        t2.join()

                        assert spy_set_experiment.call_count == 1, "set_experiment should be called once from init_tracing"

                # Verify two traces named "do" were saved to the AI System experiment
                exp = mlflow.get_experiment_by_name(expected_experiment_name)
                traces = mlflow.search_traces(
                        experiment_ids=[exp.experiment_id],
                        filter_string="trace.name = 'do'",
                        return_type='list',
                )

                # even though we don't re-initialize the experiment in both threads, the traces
                # still go to the right experiment
                assert len(traces) == 2, "Two traces named 'do' should be saved to the experiment"



import typing_extensions

from domino._impl.custommetrics.apis.tags import TagValues
from domino._impl.custommetrics.apis.tags.custom_metrics_api import CustomMetricsApi

TagToApi = typing_extensions.TypedDict(
    'TagToApi',
    {
        TagValues.CUSTOM_METRICS: CustomMetricsApi,
    }
)

tag_to_api = TagToApi(
    {
        TagValues.CUSTOM_METRICS: CustomMetricsApi,
    }
)



import time
from typing import Any, List, Optional

from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

from domino import Domino


class DominoOperator(BaseOperator):
    """
    Operator for interacting with Domino Data Lab
    via the python-domino client library w/ some
    additional functionality baked in. Follows the
    same run signature as domino.runs_start with
    some extra arguments.

    Host and API key arguments are optional and can be
    discovered via environment variable, as per the domino
    python client.

    Notes
    ------
    When combining `isDirect=True` with a command, you
    need to pass in the entire command as a single string
    in the command array.
    """

    template_fields = ("command", "title")
    ui_color = "#5188c7"

    @apply_defaults
    def __init__(
        self,
        project: str,
        command: List[str],
        host: Optional[str] = None,
        api_key: Optional[str] = None,
        domino_token_file: Optional[str] = None,
        isDirect: bool = None,
        commitId: Optional[str] = None,
        title: Optional[str] = None,
        tier: Optional[str] = None,
        publishApiEndpoint: Optional[bool] = None,
        poll_freq: Optional[int] = 5,
        max_poll_time: Optional[int] = 6000,
        startup_delay: Optional[int] = 10,
        include_setup_log: Optional[bool] = True,
        *args,
        **kwargs
    ):
        super(DominoOperator, self).__init__(*args, **kwargs)

        self.project = project
        self._api_key = api_key
        self._host = host
        self._domino_token_file = domino_token_file
        self.command = command
        self.is_direct = isDirect
        self.commit_id = commitId
        self.title = title
        self.tier = tier
        self.publish_api_endpoint = publishApiEndpoint
        self.poll_freq = poll_freq
        self.max_poll_time = max_poll_time
        self.startup_delay = startup_delay
        self.include_setup_log = include_setup_log

        self.client: Optional[Domino] = None
        self.run_id: Optional[str] = None

    def execute(self, context: Any) -> dict:

        self.log.info("Initializing Client...")
        self.client = Domino(
            self.project, self._api_key, self._host, self._domino_token_file
        )
        self.log.info("Client Initialized for project: %s", self.project)

        if self.tier:
            self.log.info("Validating Hardware Tier")
            tier_metadata = self.client.hardware_tiers_list()

            if self.tier not in [x["hardwareTier"]["name"] for x in tier_metadata]:
                raise ValueError(
                    "Provided tier not in the list of available tiers for project."
                )

        self.log.info("Starting run.")
        if self.startup_delay:
            time.sleep(self.startup_delay)

        if self.is_direct:
            if len(self.command) > 1:
                raise ValueError(
                    "Domino API will not accept a "
                    "multipart command string this long if is_direct=True"
                )

        run_response = self.client.runs_start_blocking(
            command=self.command,
            isDirect=self.is_direct,
            commitId=self.commit_id,
            title=self.title,
            tier=self.tier,
            publishApiEndpoint=self.publish_api_endpoint,
            poll_freq=self.poll_freq,
            max_poll_time=self.max_poll_time,
        )

        self.run_id = run_response["runId"]
        log = self.client.get_run_log(self.run_id, self.include_setup_log)
        self.client.process_log(log)

        return run_response


class DominoSparkOperator(BaseOperator):
    """
    Operator for launching runs via the v4 API with
    an On-Demand Spark Cluster attached.

    :param command:                             string
                                                Command to execute in Job
                                                >> domino.job_start(command="main.py arg1 arg2")
    :param commit_id:                           string (Optional)
                                                The commitId to launch from. If not provided, will launch
                                                from latest commit.
    :param hardware_tier_name:                  string (Optional)
                                                The hardware tier NAME to launch job in. If not provided
                                                it will use the default hardware tier for the project
    :param environment_id:                      string (Optional)
                                                The environment id to launch job with. If not provided
                                                it will use the default environment for the project
    :param on_demand_spark_cluster_properties:  dict (Optional)
                                                This field is deprecated. Please use compute_cluster_properties.
                                                On demand spark cluster properties. Following properties
                                                can be provided in spark cluster
                                                {
                                                    "computeEnvironmentId": "<Environment ID configured with spark>"
                                                    "executorCount": "<Number of Executors in cluster>"
                                                        (optional defaults to 1)
                                                    "executorHardwareTierId": "<Hardware tier ID for Spark Executors>"
                                                        (optional defaults to last used historically if available)
                                                    "masterHardwareTierId":  "<Hardware tier ID for Spark master"
                                                        (optional defaults to last used historically if available)
                                                    "executorStorageMB": "<Executor's storage in MB>"
                                                        (optional defaults to 0; 1GB is 1000MB Here)
                                                }
        :param compute_cluster_properties:          dict (Optional)
                                                    The compute cluster properties definition contains parameters for
                                                    launching any Domino supported compute cluster for a job. Use this
                                                    to launch a job that uses a compute cluster instead of
                                                    the deprecated on_demand_spark_cluster_properties field. If
                                                    on_demand_spark_cluster_properties and compute_cluster_properties
                                                    are both present, on_demand_spark_cluster_properties will be ignored.

                                                    compute_cluster_properties contains the following fields:
                                                    {
                                                        "clusterType": <string, one of "Ray", "Spark", "Dask", "MPI">,
                                                        "computeEnvironmentId": <string, The environment ID for the cluster's nodes>,
                                                        "computeEnvironmentRevisionSpec": <one of "ActiveRevision", "LatestRevision",
                                                        {"revisionId":"<environment_revision_id>"} (optional)>,
                                                        "masterHardwareTierId": <string, the Hardware tier ID for the cluster's master node
                                                        (required unless clusterType is MPI)>,
                                                        "workerCount": <number, the total workers to spawn for the cluster>,
                                                        "workerHardwareTierId": <string, The Hardware tier ID for the cluster workers>,
                                                        "workerStorage": <{ "value": <number>, "unit": <one of "GiB", "MB"> },
                                                        The disk storage size for the cluster's worker nodes (optional)>
                                                        "maxWorkerCount": <number, The max number of workers allowed. When
                                                        this configuration exists, autoscaling is enabled for the cluster and
                                                        'workerCount' is interpreted as the min number of workers allowed in the cluster
                                                        (optional)>
                                                    }
    :param poll_freq: int polling frequency interval in seconds
    :param max_poll_time: int max poll time in seconds
    :param startup_delay: int (Optional) number of seconds to wait before starting the job
    """

    template_fields = ("command",)
    ui_color = "#6C50AD"

    @apply_defaults
    def __init__(
        self,
        project: str,
        command: str,
        host: Optional[str] = None,
        api_key: Optional[str] = None,
        domino_token_file: Optional[str] = None,
        poll_freq: int = 5,
        max_poll_time: int = 6000,
        startup_delay: Optional[int] = None,
        commit_id: Optional[str] = None,
        hardware_tier_name: Optional[str] = None,
        environment_id: Optional[str] = None,
        on_demand_spark_cluster_properties: Optional[dict] = None,
        compute_cluster_properties: Optional[dict] = None,
        *args,
        **kwargs
    ):
        super(DominoSparkOperator, self).__init__(*args, **kwargs)

        self.log.info("Initializing Client...")

        self.project = project
        self._api_key = api_key
        self._host = host
        self._domino_token_file = domino_token_file
        self.command = command
        self.commit_id = commit_id
        self.tier = hardware_tier_name
        self.poll_freq = poll_freq
        self.max_poll_time = max_poll_time
        self.startup_delay = startup_delay
        self.environment_id = environment_id
        self.spark_propertires = on_demand_spark_cluster_properties
        self.compute_cluster_properties = compute_cluster_properties

        self.client: Optional[Domino] = None
        self.run_id: Optional[str] = None

    def execute(self, context: Any) -> dict:

        self.log.info("Initializing Client...")
        self.client = Domino(
            self.project, self._api_key, self._host, self._domino_token_file
        )
        self.log.info("Client Initialized for project: %s", self.project)

        self.log.info("Starting run.")
        if self.startup_delay:
            time.sleep(self.startup_delay)

        return self.client.job_start_blocking(
            command=self.command,
            commit_id=self.commit_id,
            hardware_tier_name=self.tier,
            environment_id=self.environment_id,
            poll_freq=self.poll_freq,
            max_poll_time=self.max_poll_time,
            on_demand_spark_cluster_properties=self.spark_propertires,
            compute_cluster_properties=self.compute_cluster_properties,
        )



from pprint import pformat, pprint

import pytest

from domino.helpers import domino_is_reachable


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_list_environments(default_domino_client):
    """
    Confirm that python-domino client can list environments (v1 API).
    """
    env_list = default_domino_client.environments_list()
    assert (
        env_list["objectType"] == "list"
    ), f"environments_list returned unexpected result:\n{pformat(env_list)}"
    assert any(
        "Domino Standard Environment" in env["name"] for env in env_list["data"]
    ), "Could not find Domino Standard Environment in environment list"


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_list_usable_environments(default_domino_client):
    """
    Confirm that python-domino client can list environments (v4 API).
    """
    env_list = default_domino_client._useable_environments_list()
    assert isinstance(
        env_list, list
    ), f"_useable_environments_list returned unexpected result:\n{pformat(env_list)}"
    assert any(
        "Domino Standard Environment" in env["name"] for env in env_list
    ), "Could not find Domino Standard Environment in environment list"


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_spark_in_environments(default_domino_client):
    """
    Confirm that python-domino client can list environments (v1 API).
    """
    env_list = default_domino_client.environments_list()
    assert (
        env_list["objectType"] == "list"
    ), f"environments_list returned unexpected result:\n{pformat(env_list)}"

    for env in env_list["data"]:
        if "Spark Cluster Environment" in env["name"]:
            pprint(env)

    assert any(
        "Spark Cluster Environment" in env["name"] for env in env_list["data"]
    ), "Could not find Domino Standard Environment in environment list"




import os

import docker
import pytest
import requests_mock
from requests.auth import AuthBase

from domino import Domino
from domino.constants import (
    DOMINO_HOST_KEY_NAME,
    DOMINO_TOKEN_FILE_KEY_NAME,
    DOMINO_USER_API_KEY_KEY_NAME,
    DOMINO_USER_NAME_KEY_NAME,
)

TEST_AI_SYSTEMS_ENV_VARS = {
    "MLFLOW_TRACKING_URI": "http://localhost:5000",
    "DOMINO_AI_SYSTEM_CONFIG_PATH": "tests/assets/ai_system_config.yaml",
    "DOMINO_AI_SYSTEM_IS_PROD": "false"
}

version_info = {
    "buildId": "12345",
    "buildUrl": "https://server.somefakecompany.com/domino/12345",
    "commitId": "123deadbeef456",
    "commitUrl": "https://repo.somefakecompany.com/domino/commit/123deadbeef456",
    "timestamp": "2021-06-14T16:50:02Z",
    "version": "9.9.9",
}

@pytest.fixture
def dummy_hostname():
    return "http://domino.somefakecompany.com"


@pytest.fixture(scope="module")
def default_domino_client():
    """
    Module-scoped fixture that uses default authentiation, and provides a Domino client object.
    """
    host = os.getenv(DOMINO_HOST_KEY_NAME)
    user = os.getenv(DOMINO_USER_NAME_KEY_NAME)
    assert (
        user
    ), f"User must be specified by {DOMINO_USER_NAME_KEY_NAME} environment variable."
    assert (
        host
    ), f"Host must be specified by {DOMINO_HOST_KEY_NAME} environment variable."

    domino_client = Domino(host=host, project=f"{user}/quick-start")

    # If authentication failed, this raises a requests.exceptions.HTTPError.
    domino_client.environments_list()

    return domino_client


@pytest.fixture(scope="module")
def spark_compute_env_id():
    """
    Module-scoped fixture that uses default authentication, and provides a Domino client object.
    """
    host = os.getenv(DOMINO_HOST_KEY_NAME)
    user = os.getenv(DOMINO_USER_NAME_KEY_NAME)
    assert (
        user
    ), f"User must be specified by {DOMINO_USER_NAME_KEY_NAME} environment variable."
    assert (
        host
    ), f"Host must be specified by {DOMINO_HOST_KEY_NAME} environment variable."

    domino_client = Domino(host=host, project=f"{user}/quick-start")

    env_list = domino_client.environments_list()

    for env in env_list["data"]:
        if "Spark Compute Environment" in env["name"]:
            return env["id"]


@pytest.fixture(scope="module")
def spark_cluster_env_id():
    """
    Module-scoped fixture that uses default authentication, and provides a Domino client object.
    """
    host = os.getenv(DOMINO_HOST_KEY_NAME)
    user = os.getenv(DOMINO_USER_NAME_KEY_NAME)
    assert (
        user
    ), f"User must be specified by {DOMINO_USER_NAME_KEY_NAME} environment variable."
    assert (
        host
    ), f"Host must be specified by {DOMINO_HOST_KEY_NAME} environment variable."

    domino_client = Domino(host=host, project=f"{user}/quick-start")

    env_list = domino_client.environments_list()

    for env in env_list["data"]:
        if "Spark Cluster Environment" in env["name"]:
            return env["id"]


@pytest.fixture
def mock_proxy_response_https():
    """
    Simulate a valid response to an authenticated /version GET request using api proxy with schema
    """

    dummy_url = "https://localhost:1234"
    with requests_mock.Mocker() as mock_endpoint:
        response = mock_endpoint.get(
            f"{dummy_url}/version", json=version_info, status_code=200
        )
        yield response


@pytest.fixture
def mock_proxy_response():
    """
    Simulate a valid response to an authenticated /version GET request using api proxy
    """

    dummy_url = "http://localhost:1234"
    with requests_mock.Mocker() as mock_endpoint:
        response = mock_endpoint.get(
            f"{dummy_url}/version", json=version_info, status_code=200
        )
        yield response



@pytest.fixture
def mock_domino_version_response():
    """
    Simulate a valid response to an authenticated /version GET request.

    Assumes the target endpoint is domino.somefakecompany.com/version.
    """

    dummy_url = "http://domino.somefakecompany.com"
    with requests_mock.Mocker() as mock_endpoint:
        response = mock_endpoint.get(
            f"{dummy_url}/version", json=version_info, status_code=200
        )
        yield response


@pytest.fixture
def dummy_token_file(tmpdir):
    """
    Simulate a token file.
    """
    # Refer to https://docs.pytest.org/en/6.2.x/tmpdir.html#the-tmpdir-fixture for tmpdir usage
    token_file = tmpdir.join("dummy_token_file.txt")
    token_file.write("top_secret_auth_token")
    return token_file


@pytest.fixture
def clear_token_file_from_env():
    """
    Unset any DOMINO_TOKEN_FILE var from the environment for API key tests.

    If a token file is present in the environment, python-domino uses it preferentially,
    which is mutually incompatible with testing the API key.
    """
    saved_environment = dict(os.environ)
    os.environ.pop(DOMINO_TOKEN_FILE_KEY_NAME, default=None)
    yield

    # Restore original pre-test environment
    os.environ.update(saved_environment)


@pytest.fixture
def clear_api_key_from_env():
    """
    Unset any DOMINO_USER_API_KEY var from the environment.

    If the API key is present in the environment, python-domino uses it automatically,
    which will affect some negative testing.
    """
    saved_environment = dict(os.environ)
    os.environ.pop(DOMINO_USER_API_KEY_KEY_NAME, default=None)
    yield

    # Restore original pre-test environment
    os.environ.update(saved_environment)


@pytest.fixture
def test_auth_base():
    class TestAuth(AuthBase):
        def __init__(self, *args, **kwargs):
            super(TestAuth, self).__init__(*args, **kwargs)
            self.header = None
    return TestAuth()

@pytest.fixture(scope="session")
def docker_client():
    return docker.from_env()



import time
from pprint import pformat

import polling2
import pytest
from requests.exceptions import RequestException

from domino import Domino
from domino.helpers import domino_is_reachable


@pytest.fixture
def mock_job_start_blocking_setup(requests_mock, dummy_hostname):
    """
    Some of the tests in this file are true unit tests, and do not require
    a live deployment. In order for them to run, any API calls made by
    job_start() internally need to have mocked responses.

    This is the test fixture where all of the mocked API return values are
    created.

    If any dependent calls to the API are added to job_start, then they must
    be mocked here as well.
    """
    # Mock the /version API endpoint (GET)
    requests_mock.get(f"{dummy_hostname}/version", json={"version": "9.9.9"})

    # Mock /findProjectByOwnerAndName API endpoint (GET) and return json with ID abcdef
    project_endpoint = "v4/gateway/projects/findProjectByOwnerAndName"
    project_query = "ownerName=anyuser&projectName=anyproject"
    requests_mock.get(
        f"{dummy_hostname}/{project_endpoint}?{project_query}", json={"id": "abcdef"}
    )

    # Mock the jobs/start API endpoint (POST) and return json with ID 123
    jobs_start_endpoint = "v4/jobs/start"
    requests_mock.post(f"{dummy_hostname}/{jobs_start_endpoint}", json={"id": "123"})

    # Mock STDOUT for run with ID 123 add return json with string value
    stdout_endpoint = "v1/projects/anyuser/anyproject/run/123/stdout"
    requests_mock.get(
        f"{dummy_hostname}/{stdout_endpoint}", json={"stdout": "whatever"}
    )

    # Mock HWT endpoint
    hwt_endpoint = "v4/projects/abcdef/hardwareTiers"
    requests_mock.get(f"{dummy_hostname}/{hwt_endpoint}", json=[])
    yield


@pytest.mark.usefixtures("clear_token_file_from_env", "mock_job_start_blocking_setup")
def test_job_status_completes_with_default_params(requests_mock, dummy_hostname):
    """
    Confirm that the happy path default case passes (no exceptions thrown)
    """
    # Mock a typical response from the jobs status API endpoint (GET)
    jobs_status_endpoint = "v4/jobs/123"
    requests_mock.get(
        f"{dummy_hostname}/{jobs_status_endpoint}",
        json={"statuses": {"isCompleted": True, "executionStatus": "whatever"}},
    )

    d = Domino(host=dummy_hostname, project="anyuser/anyproject", api_key="whatever")

    job_status = d.job_start_blocking(command="foo.py", poll_freq=1, max_poll_time=1)
    assert job_status["statuses"]["isCompleted"] is True


@pytest.mark.usefixtures("clear_token_file_from_env", "mock_job_start_blocking_setup")
def test_job_status_ignores_RequestException_and_times_out(
    requests_mock, dummy_hostname
):
    """
    Test that the default behavior is to simply ignore RequestException being thrown.
    (In this case, timing out via polling2.TimeoutException is expected.)
    """
    # Force the jobs status API endpoint to throw a RequestException when called
    jobs_status_endpoint = "v4/jobs/123"
    requests_mock.get(f"{dummy_hostname}/{jobs_status_endpoint}", exc=RequestException)

    d = Domino(host=dummy_hostname, project="anyuser/anyproject", api_key="whatever")

    with pytest.raises(polling2.TimeoutException):
        d.job_start_blocking(command="foo.py", poll_freq=1, max_poll_time=1)


@pytest.mark.usefixtures("clear_token_file_from_env", "mock_job_start_blocking_setup")
def test_job_status_without_ignoring_exceptions(requests_mock, dummy_hostname):
    """
    Test that ignore_exceptions can be overridden by passing in an empty tuple.
    The call should fail immediately with RequestException.
    """
    # Force the jobs status API endpoint to throw a RequestException when called
    jobs_status_endpoint = "v4/jobs/123"
    requests_mock.get(f"{dummy_hostname}/{jobs_status_endpoint}", exc=RequestException)

    d = Domino(host=dummy_hostname, project="anyuser/anyproject", api_key="whatever")

    with pytest.raises(RequestException):
        d.job_start_blocking(command="foo.py", ignore_exceptions=())


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_job_start_blocking(default_domino_client):
    """
    Confirm that we can start a job using the v4 API, and block until it succeeds.
    """
    job = default_domino_client.job_start_blocking(command="main.py")
    assert job["statuses"]["executionStatus"] == "Succeeded"


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_job_start_override_hardware_tier_id(default_domino_client):
    """
    Confirm that we can start a job using the v4 API and override the hardware tier id
    """
    hardware_tiers = default_domino_client.hardware_tiers_list()
    non_default_hardware_tiers = [
        hwt for hwt in hardware_tiers if not hwt["hardwareTier"]["hwtFlags"]["isDefault"]
    ]
    if len(non_default_hardware_tiers) == 0:
        pytest.xfail("No non-default hardware tiers found: cannot run test")

    override_hardware_tier_id = non_default_hardware_tiers[0]["hardwareTier"]["id"]
    job_status = default_domino_client.job_start_blocking(
        command="main.py", hardware_tier_id=override_hardware_tier_id
    )
    assert job_status["statuses"]["isCompleted"] is True
    job_red = default_domino_client.job_runtime_execution_details(job_status["id"])
    assert job_red["hardwareTier"]["id"] == override_hardware_tier_id


# deprecated but ensuring it still works for now
@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_job_start_override_hardware_tier_name(default_domino_client):
    """
    Confirm that we can start a job using the v4 API and override the hardware tier via hardware_tier_name
    """
    hardware_tiers = default_domino_client.hardware_tiers_list()

    non_default_hardware_tiers = [
        hwt for hwt in hardware_tiers if not hwt["hardwareTier"]["hwtFlags"]["isDefault"]
    ]
    if len(non_default_hardware_tiers) == 0:
        pytest.xfail("No non-default hardware tiers found: cannot run test")

    override_hardware_tier_name = non_default_hardware_tiers[0]["hardwareTier"]["name"]
    job_status = default_domino_client.job_start_blocking(
        command="main.py", hardware_tier_name=override_hardware_tier_name
    )

    assert job_status["statuses"]["isCompleted"] is True
    job_red = default_domino_client.job_runtime_execution_details(job_status["id"])
    assert job_red["hardwareTier"]["name"] == override_hardware_tier_name


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_runs_list(default_domino_client):
    """
    Confirm that the v1 API endpoint to list jobs returns a list.
    """
    runs = default_domino_client.runs_list()
    assert (
        runs["objectType"] == "list"
    ), f"runs_list returned unexpected result:\n{pformat(runs)}"
    assert isinstance(
        runs["data"], list
    ), f"runs_list returned unexpected result:\n{pformat(runs)}"


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_queue_job(default_domino_client):
    """
    Queue a job, and then poll until the job completes (timeout at 240 seconds).
    """
    job = default_domino_client.job_start("main.py")

    remaining_polling_seconds = 240
    while remaining_polling_seconds > 0:
        job_status = default_domino_client.job_status(job["id"])
        if not job_status["statuses"]["isCompleted"]:
            time.sleep(3)
            remaining_polling_seconds -= 3
            print(f"Job {job['id']} has not completed.")
            continue

        exec_status = job_status["statuses"]["executionStatus"]
        try:
            assert exec_status == "Succeeded"
            print(f"Job {job['id']} succeeded.")
            break
        except AssertionError:
            print(f"Job failed: {pformat(job_status)}")
            raise
    else:
        pytest.fail(f"Job took too long to complete: {pformat(job_status)}")


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_job_result_stdout(default_domino_client):
    """
    Queue a job, and then poll until the job completes (timeout at 240 seconds).
    """
    job = default_domino_client.job_start("main.py")

    remaining_polling_seconds = 240
    while remaining_polling_seconds > 0:
        job_status = default_domino_client.job_status(job["id"])
        if not job_status["statuses"]["isCompleted"]:
            time.sleep(5)
            remaining_polling_seconds -= 5
            continue

        stdout_data = default_domino_client.runs_stdout(job["id"])

        try:
            html_start_tags = "<pre style='white-space: pre-wrap; white-space"
            assert stdout_data != ""
            assert html_start_tags not in stdout_data
            break
        except AssertionError:
            print(f"Job failed: {pformat(job_status)}")
            raise
    else:
        pytest.fail(f"Job took too long to complete: {pformat(job_status)}")



import logging
import mlflow
import threading
from typing import Optional

from .._client import client
from .._constants import EXPERIMENT_AI_SYSTEM_TAG
from ._util import is_ai_system, get_running_ai_system_experiment_name
from .._verify_domino_support import verify_domino_support

# init_tracing is not thread safe. this likely won't cause an issue with the autolog frameworks. If data inconsistency is caused with
# autolog frameworks, then the worst case scenario is that we get duplicate autolog calls. These are local to the process
# so not a big deal

global _triggered_autolog_frameworks
_triggered_autolog_frameworks = set()

global _is_prod_tracing_initialized
_is_prod_tracing_initialized = False

# Lock to ensure thread-safe access to _is_prod_tracing_initialized
global _prod_tracing_init_lock
_prod_tracing_init_lock = threading.Lock()


def init_tracing(autolog_frameworks: Optional[list[str]] = None):
    """Initialize Mlflow autologging for various frameworks and sets the active experiment to enable tracing in production.
    This may be used to initialize logging and tracing for the AI System in dev and prod modes.

    In prod mode, environment variables DOMINO_AI_SYSTEM_IS_PROD, DOMINO_APP_ID
    must be set. Call init_tracing before your app starts up to start logging traces to Domino.

    Args:
        autolog_frameworks: list of frameworks to autolog
    """
    verify_domino_support()
    frameworks = autolog_frameworks or []
    global _is_prod_tracing_initialized
    # Guard and initialization are protected by a lock for thread safety
    if is_ai_system():
        with _prod_tracing_init_lock:
            if not _is_prod_tracing_initialized:
                # activate ai system experiment
                logger.debug("Initializing mlflow tracing for AI System")

                # we use the client to create experiment and get to avoid racy behavior
                experiment = client.get_experiment_by_name(
                    get_running_ai_system_experiment_name()
                )
                if experiment is None:
                    experiment_name = get_running_ai_system_experiment_name()

                    logger.debug(
                        "Creating new experiment for AI System named %s",
                        experiment_name,
                    )

                    experiment_id = client.create_experiment(experiment_name)

                    logger.debug(
                        "Created experiment for AI System with ID %s", experiment_id
                    )

                    client.set_experiment_tag(
                        experiment_id, EXPERIMENT_AI_SYSTEM_TAG, "true"
                    )

                    logger.debug(
                        "Tagged experiment with ID %s with tag %s",
                        experiment_id,
                        EXPERIMENT_AI_SYSTEM_TAG,
                    )
                else:
                    experiment_id = experiment.experiment_id

                # only set experiment by ID to avoid the python client creating a new experiment with a random ID appended
                # this happens when init_tracing called by itself and then a domino trace started right after

                logger.debug(
                    "Setting AI System experiment with ID %s as active", experiment_id
                )

                experiment = mlflow.set_experiment(experiment_id=experiment_id)

                _is_prod_tracing_initialized = True

    for fw in frameworks:
        global _triggered_autolog_frameworks
        if fw not in _triggered_autolog_frameworks:
            _triggered_autolog_frameworks.add(fw)
            call_autolog(fw)


def call_autolog(fw: str):
    try:
        getattr(mlflow, fw).autolog()
    except Exception as e:
        logger.warning(f"Failed to call mlflow autolog for {fw} ai framework", e)


logger = logging.getLogger(__name__)



# TODO update tag names to be prefixed with domino.aisystems in order to make origin clear
EVALUATION_TAG_PREFIX = "domino.prog"
DOMINO_INTERNAL_EVAL_TAG = "domino.internal.is_eval"
MIN_MLFLOW_VERSION = "3.2.0"
MIN_DOMINO_VERSION = "6.2.0"
LARGEST_MAX_RESULTS_PAGE_SIZE = 500
EXPERIMENT_AI_SYSTEM_TAG = "domino.ai_system"



# do not import all endpoints into this module because that uses a lot of memory and stack frames
# if you need the ability to import all endpoints from this module, import them with
# from domino._impl.custommetrics.paths.api_metric_alerts_v1 import Api

from domino._impl.custommetrics.paths import PathValues

path = PathValues.API_METRIC_ALERTS_V1


from dataclasses import dataclass
from datetime import datetime
import functools
import inspect
import logging
import mlflow
from typing import Optional, Callable, Any
from uuid import uuid4

from .._client import client
from .inittracing import init_tracing
from ..logging.logging import log_evaluation
from ._util import get_is_production, build_ai_system_experiment_name
from .._eval_tags import validate_label, get_eval_tag_name
from .._verify_domino_support import verify_domino_support

EvalResult = dict[str, int | float | str]

SpanEvaluator = Callable[[mlflow.entities.Span], EvalResult]
TraceEvaluator = Callable[[mlflow.entities.Trace], EvalResult]

DOMINO_NO_RESULT_ADD_TRACING = "domino_no_result"

@dataclass
class SpanSummary:
    """A span in a trace."""

    id: str
    """the mlflow ID of the span"""

    name: str
    """The name of the span"""

    trace_id: str
    """The parent trace ID"""

    inputs: Any
    """The inputs to the function that created the span"""

    outputs: Any
    """The outputs of the function that created the span"""

@dataclass
class EvaluationResult:
    """An evaluation result for a trace."""

    name: str
    """The name of the evaluation"""

    value: float | str
    """The value of the evaluation"""


@dataclass
class TraceSummary:
    """A summary of a trace."""

    name: str
    """The name of the trace"""

    id: str
    """The mlflow ID of the trace"""

    spans: list[SpanSummary]
    """The child spans of this trace"""

    evaluation_results: list[EvaluationResult]
    """The evaluation results for this trace"""


@dataclass
class SearchTracesResponse:
    """The response from searching for traces."""

    data: list[TraceSummary]
    """The list of trace summaries"""

    page_token: Optional[str]
    """The token for the next page of results"""


def _datetime_to_ms(dt: datetime) -> int:
    return dt.timestamp() * 1000


def _make_span_summary(span: mlflow.entities.Span) -> SpanSummary:
    return SpanSummary(
        id=span.span_id,
        name=span.name,
        trace_id=span.trace_id,
        inputs=span.inputs,
        outputs=span.outputs,
    )


def _do_evaluation(
    span: mlflow.entities.Span,
    evaluator: Optional[SpanEvaluator] = None,
    trace_evaluator: Optional[TraceEvaluator] = None,
    is_production: bool = False,
) -> Optional[dict]:
    if not is_production and evaluator or trace_evaluator:
        span_eval_result = None
        trace_eval_result = None

        # get span evaluator result
        if evaluator:
            try:
                span_eval_result = evaluator(span)
            except Exception as e:
                logger.error(
                    "Inline evaluation failed for evaluator, %s. Error: %s",
                    evaluator.__name__,
                    e,
                    exc_info=True,
                )

        # get trace evaluator result
        if trace_evaluator:
            try:
                trace = mlflow.get_trace(span.trace_id)
            except Exception as e:
                logger.warning(
                    "A trace_evaluator %s was provided, but the trace could not be found: %s",
                    trace_evaluator.__name__,
                    e,
                )
                trace = None

            if trace:
                try:
                    trace_eval_result = trace_evaluator(trace)
                except Exception as e:
                    logger.error(
                        "Inline evaluation failed for trace_evaluator, %s. Error: %s",
                        trace_evaluator.__name__,
                        e,
                        exc_info=True,
                    )

            else:
                # warn the user if they provided a trace evaluator, but we couldn't find a trace
                # because they may need to design their evaluations differently
                logger.warning(
                    "A trace_evaluator %s was provided, but the trace could not be found",
                    trace_evaluator.__name__,
                )

        # if at least one result exists, return a combined dict
        # otherwise, return none
        if span_eval_result or trace_eval_result:
            return {**(trace_eval_result or {}), **(span_eval_result or {})}

    return None


def _log_eval_results(
    parent_span: mlflow.entities.Span,
    evaluator: Optional[SpanEvaluator],
    trace_evaluator: Optional[TraceEvaluator],
):
    """
    Saves the evaluation results
    """
    is_production = get_is_production()
    eval_result = _do_evaluation(parent_span, evaluator, trace_evaluator, is_production)

    if eval_result:
        for k, v in eval_result.items():
            # save the evaluation result for the trace
            log_evaluation(
                parent_span.request_id,
                value=v,
                name=k,
            )


def _set_span_inputs(parent_span, func, args, kwargs):
    """
    Sets the inputs on the span. Takes the arguments passed to the decorated function
    and preserves names of positional arguments, ensures defaults are applied, and excludes
    'self' and 'cls'  in the case of class methods and instance methods, because the user doesn't
    explicitly pass these in via the args and we want the arguments to match what the user expects
    """
    bound_args = inspect.signature(func).bind(*args, **kwargs)
    bound_args.apply_defaults()
    inputs = {
        k: v for k, v in bound_args.arguments.items() if k != "self" and k != "cls"
    }
    parent_span.set_inputs(inputs)

    return inputs


def add_tracing(
    name: str,
    autolog_frameworks: Optional[list[str]] = [],
    evaluator: Optional[SpanEvaluator] = None,
    trace_evaluator: Optional[TraceEvaluator] = None,
    eagerly_evaluate_streamed_results: bool = True,
):
    """This is a decorator that starts an mlflow span for the function it decorates. If there is an existing trace
    a span will be appended to it. If there is no existing trace, a new trace will be created.

    It also enables the user to run evaluators when the code is run in development mode. Evaluators can be run on
    the span and/or trace generated for the wrapped function call. The trace evaluator will run if the parent trace
    was started and finished by the related decorator call. The trace will contain all child span information.
    The span evaluator will always run. The evaluation results from both evaluators will be combined and saved
    to the trace.

    This decorator must be used directly on the function to be traced without any intervening decorators, because it
    must have access to the arguments.

    @add_tracing(
        name="assistant_chat_bot",
        evaluator=evaluate_helpfulness,
    )
    def ask_chat_bot(user_input: str) -> dict:
        ...

    Args:
        name: the name of the span to add to existing trace or create if no trace exists yet.

        autolog_frameworks: an optional list of mlflow supported frameworks to autolog

        evaluator: an optional function that takes the span created for the wrapped function and returns a dictionary of evaluation results. The evaluation results will be saved to the trace

        trace_evaluator: an optional function that takes the trace for this call stack and returns a dictionary of evaluation results. This evaluator will be triggered if the trace was started and finished by the add tracing decorator. The evaluation results will be saved to the trace

        eagerly_evaluate_streamed_results: optional boolean, defaults to true, this determines if all
            yielded values should be aggregated and set as outputs to a single span. This makes evaluation easier, but
            will impact performance if you expect a large number of streamed values. If set to false, each yielded value
            will generate a new span on the trace, which can be evaluated post-hoc. Inline evaluators won't be executed.
            Each span will have a group_id set in their attributes to indicate that they are part of the same function call.
            Each span will have an index to indicate what order they arrived in.

    Returns:
        A decorator that wraps the function to be traced.
    """
    validate_label(name)


    def decorator(func):
        # For Regular Functions (e.g., langgraph_agents.run_agent)
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            result = DOMINO_NO_RESULT_ADD_TRACING
            init_tracing(autolog_frameworks)

            with mlflow.start_span(name) as parent_span:
                _set_span_inputs(parent_span, func, args, kwargs)

                result = func(*args, **kwargs)

                parent_span.set_outputs(result)

            if result != DOMINO_NO_RESULT_ADD_TRACING:
                _log_eval_results(parent_span, evaluator, trace_evaluator)

            return _return_traced_result(result)

        # for async functions
        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def async_wrapper(*args, **kwargs):
                result = DOMINO_NO_RESULT_ADD_TRACING
                init_tracing(autolog_frameworks)

                with mlflow.start_span(name) as parent_span:
                    _set_span_inputs(parent_span, func, args, kwargs)

                    result = await func(*args, **kwargs)

                    parent_span.set_outputs(result)

                if result != DOMINO_NO_RESULT_ADD_TRACING:
                    _log_eval_results(parent_span, evaluator, trace_evaluator)

                return _return_traced_result(result)

            return async_wrapper

        # For Generator Functions (e.g., RAGAgent.answer)
        if inspect.isgeneratorfunction(func):

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                result = DOMINO_NO_RESULT_ADD_TRACING
                init_tracing(autolog_frameworks)

                with mlflow.start_span(name) as parent_span:
                    inputs = _set_span_inputs(parent_span, func, args, kwargs)

                    result = func(*args, **kwargs)

                    if not eagerly_evaluate_streamed_results:
                        # can't do inline evaluation, so warn if an evaluator is provided
                        if evaluator or trace_evaluator:
                            logger.warning(
                                """eagerly_evaluate_streamed_results is false, so inline evaluation is disabled.
                                You can still do evaluations using log_evaluation post-hoc"""
                            )

                        group_id = uuid4()
                        i = -1
                        for v in result:
                            i += 1
                            with mlflow.start_span(name) as gen_span:
                                # make span for each yielded value
                                gen_span.set_inputs(inputs)
                                gen_span.set_attributes(
                                    {"group_id": str(group_id), "index": i}
                                )
                                yield v
                                gen_span.set_outputs(v)
                    else:
                        # get all results and set as outputs
                        all_results = [v for v in result]
                        parent_span.set_outputs(all_results)
                        for v in all_results:
                            yield v

                if eagerly_evaluate_streamed_results and result != DOMINO_NO_RESULT_ADD_TRACING:
                    _log_eval_results(parent_span, evaluator, trace_evaluator)

            return wrapper

        return wrapper

    return decorator


def _build_evaluation_result(tag_key: str, tag_value: str) -> EvaluationResult:
    value = tag_value
    try:
        value = float(tag_value)
    except Exception:
        pass

    return EvaluationResult(name=get_eval_tag_name(tag_key), value=value)


def _build_trace_summaries(traces) -> list[TraceSummary]:
    trace_summaries = []
    for trace in traces:
        # build trace summaries if there is a span location for the trace
        parent_span = trace.data.spans[0]
        trace_name = parent_span.name
        spans = trace.data.spans

        trace_summaries.append(
            TraceSummary(
                name=trace_name,
                id=trace.info.trace_id,
                spans=[_make_span_summary(s) for s in spans],
                evaluation_results=[
                    _build_evaluation_result(key, value)
                    for (key, value) in trace.info.tags.items()
                    if get_eval_tag_name(key) is not None
                ],
            )
        )

    return trace_summaries


def search_ai_system_traces(
    ai_system_id: str,
    ai_system_version: Optional[str] = None,
    trace_name: Optional[str] = None,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
    page_token: Optional[str] = None,
    max_results: Optional[int] = None,
) -> SearchTracesResponse:
    """This allows searching for traces that have a certain name and returns a paginated response of trace summaries that
    include the spans that were requested.

    Args:
        ai_system_id: string, the ID of the AI System to filter by
        ai_system_version: string, the version of the AI System to filter by, if not provided will search throuh all versions
        trace_name: the name of the traces to search for
        start_time: python datetime
        end_time: python datetime, defaults to now
        page_token: page token for pagination. You can use this to request the next page of results and may
         find a page_token in the response of the previous search_traces call.
        max_results: defaults to 100

    Returns:
        SearchTracesResponse: a token based pagination response that contains a list of trace summaries
            data: list of TraceSummary
            page_token: the next page's token
    """
    return _search_traces(
        None,
        ai_system_id,
        ai_system_version,
        trace_name,
        start_time,
        end_time,
        page_token,
        max_results,
    )


def search_traces(
    run_id: str,
    trace_name: Optional[str] = None,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
    page_token: Optional[str] = None,
    max_results: Optional[int] = None,
) -> SearchTracesResponse:
    """This allows searching for traces that have a certain name and returns a paginated response of trace summaries that
    inclued the spans that were requested.

    Args:
        run_id: string, the ID of the development mode evaluation run to search for traces.
        trace_name: the name of the traces to search for
        start_time: python datetime
        end_time: python datetime, defaults to now
        page_token: page token for pagination. You can use this to request the next page of results and may
         find a page_token in the response of the previous search_traces call.
        max_results: defaults to 100

    Returns:
        SearchTracesReponse: a token based pagination response that contains a list of trace summaries
            data: list of TraceSummary
            page_token: the next page's token
    """
    return _search_traces(
        run_id,
        None,
        None,
        trace_name,
        start_time,
        end_time,
        page_token,
        max_results,
    )


def _search_traces(
    run_id: Optional[str] = None,
    ai_system_id: Optional[str] = None,
    ai_system_version: Optional[str] = None,
    trace_name: Optional[str] = None,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
    page_token: Optional[str] = None,
    max_results: Optional[int] = None,
) -> SearchTracesResponse:
    # this depends on mlflow 3 due to the pagination support
    verify_domino_support()

    if not run_id and not ai_system_version and not ai_system_id:
        raise Exception(
            "Either run_id or ai_system_id and ai_system_version must be provided to search traces"
        )

    if ai_system_version and not ai_system_id:
        raise Exception(
            "If ai_system_version is provided, ai_system_id must also be provided"
        )

    filter_clauses = []

    # get experiment id
    if run_id:
        experiment_id = client.get_run(run_id).info.experiment_id
        run_filter_clause = f'metadata.mlflow.sourceRun = "{run_id}"'
        filter_clauses.append(run_filter_clause)
    else:
        experiment_name = build_ai_system_experiment_name(ai_system_id)
        experiment = client.get_experiment_by_name(experiment_name)
        if not experiment:
            raise Exception(
                f"No experiment found for ai_system_id: {ai_system_id} and ai_system_version: {ai_system_version}"
            )
        experiment_id = experiment.experiment_id

        # add prod trace tag filters
        if ai_system_id:
            filter_clauses.append(f'tags.mlflow.domino.app_id = "{ai_system_id}"')

        if ai_system_version:
            filter_clauses.append(
                f'tags.mlflow.domino.app_version_id = "{ai_system_version}"'
            )

    if start_time or end_time:
        time_range_filter_clause = ""

        if start_time:
            start_ms = _datetime_to_ms(start_time)
            time_range_filter_clause += f"timestamp_ms > {start_ms}"

        if end_time:
            end_ms = _datetime_to_ms(end_time)
            time_range_filter_clause += f" AND timestamp_ms < {end_ms}"

        if start_time and end_time and start_time >= end_time:
            logger.warning("start_time must be before end_time")

        filter_clauses.append(time_range_filter_clause)

    if trace_name:
        trace_name_filter_clause = f'trace.name = "{trace_name}"'
        filter_clauses.append(trace_name_filter_clause)

    # get traces made within the time range
    # get traces with the trace names
    filter_string = " AND ".join([x for x in filter_clauses if x])

    traces = client.search_traces(
        experiment_ids=[experiment_id],
        filter_string=filter_string,
        page_token=page_token,
        max_results=max_results or 100,
        order_by=["attributes.timestamp_ms ASC"],
    )

    trace_summaries = _build_trace_summaries(traces)
    next_page_token = traces.token

    return SearchTracesResponse(trace_summaries, next_page_token)

def _return_traced_result(result: any):
    if result != DOMINO_NO_RESULT_ADD_TRACING:
        return result
    else:
        logger.warning("No result returned from traced function")

logger = logging.getLogger(__name__)



import warnings

from urllib.parse import quote

from typing import Optional


class _Routes:
    def __init__(self, host, owner_username, project_name):
        self.host = host
        self._owner_username = owner_username
        self._project_name = project_name

    # URL builders
    def _build_project_url(self):
        return (
            self.host
            + "/v1/projects/"
            + self._owner_username
            + "/"
            + self._project_name
        )

    def _build_project_url_private_api(self):
        return self.host + "/u/" + self._owner_username + "/" + self._project_name

    def _build_old_project_url(self):
        # TODO refactor once these API endpoints are supported in REST API
        return self.host + "/" + self._owner_username + "/" + self._project_name

    def _build_models_url(self):
        return self.host + "/v1/models"

    def _build_models_v4_url(self):
        return self.host + "/v4/models"

    # Project URLs
    def project_create(self):
        return self.host + "/project"

    def project_v4(self, project_id: Optional[str] = None) -> str:
        return self.host + "/v4/projects" + (f"/{project_id}" if project_id else "")

    def projects_list(self):
        return f"{self.host}/v4/gateway/projects"

    def tag_details(self, tag_id):
        return f"{self.host}/projectTags/{tag_id}"

    def tags_add(self, project_id):
        return f"{self.host}/v4/projects/{project_id}/tags"

    def tags_remove(self, project_id, tag_id):
        return f"{self.host}/v4/projects/{project_id}/tags/{tag_id}"

    def runs_list(self):
        return self._build_project_url() + "/runs"

    def runs_start(self):
        return self._build_project_url() + "/runs"

    def runs_status(self, runId):
        return self._build_project_url() + "/runs/" + runId

    def runs_stdout(self, runId):
        return self._build_project_url() + "/run/" + runId + "/stdout"

    def files_list(self, commitId, path):
        return self._build_project_url() + "/files/" + commitId + "/" + path

    def files_upload(self, path):
        return self._build_project_url() + path

    def commits_list(self):
        return self._build_project_url() + "/commits"

    # Deprecated - use blobs_get_v2 instead
    def blobs_get(self, key):
        message = "blobs_get is deprecated and will soon be removed. Please migrate to blobs_get_v2 and adjust the " \
                  "input parameters accordingly "
        warnings.warn(message, DeprecationWarning)
        return self._build_project_url() + "/blobs/" + key

    def blobs_get_v2(self, path, commit_id, project_id):
        return self.host + f"/api/projects/v1/projects/{project_id}/files/{commit_id}/{path}/content"

    def fork_project(self, project_id):
        return self.host + f"/v4/projects/{project_id}/fork"

    def collaborators_get(self):
        return self._build_old_project_url() + "/collaborators"

    def collaborators_add(self, project_id):
        return self.host + f"/v4/projects/{project_id}/collaborators"

    def collaborators_remove(self, project_id, user_id):
        return self.host + f"/v4/projects/{project_id}/collaborators/{user_id}"

    # API Endpoint URLs
    def _build_endpoint_url(self):
        return (
            self.host
            + "/v1/"
            + self._owner_username
            + "/"
            + self._project_name
            + "/endpoint"
        )

    def endpoint(self):
        return self._build_endpoint_url()

    def endpoint_state(self):
        return self._build_endpoint_url() + "/state"

    def endpoint_publish(self):
        return self._build_endpoint_url() + "/publishRelease"

    # Model Manager URLs
    def models_list(self):
        return self._build_project_url() + "/models"

    def model_publish(self):
        return self._build_models_url()

    def model_versions_get(self, model_id):
        return self._build_models_url() + "/" + model_id + "/versions"

    def model_version_publish(self, model_id):
        return self._build_models_url() + "/" + model_id + "/versions"

    def model_version_export(self, model_id, model_version_id):
        return (
            self._build_models_v4_url()
            + "/"
            + model_id
            + "/"
            + model_version_id
            + "/exportImageToRegistry"
        )

    def model_version_sagemaker_export(self, model_id, model_version_id):
        return (
            self._build_models_v4_url()
            + "/"
            + model_id
            + "/"
            + model_version_id
            + "/exportImageForSagemaker"
        )

    def model_version_export_status(self, model_export_id):
        return (
            self._build_models_v4_url()
            + "/"
            + model_export_id
            + "/getExportImageStatus"
        )

    def model_version_export_logs(self, model_export_id):
        return self._build_models_v4_url() + "/" + model_export_id + "/getExportLogs"

    # Environment URLs
    def _build_v4_environments_url(self) -> str:
        return self.host + "/v4/environments"

    def _build_v1_environments_url(self) -> str:
        return self.host + "/api/environments/v1/environments"

    def _build_beta_environments_url(self) -> str:
        return self.host + "/api/environments/beta/environments"

    def environments_list(self):
        return self.host + "/v1/environments"

    def environment_default_get(self):
        return self._build_v4_environments_url() + "/defaultEnvironment"

    def environment_default_set(self):
        return self._build_v4_environments_url() + "/setDefaultEnvironment"

    def environment_create(self):
        return self._build_beta_environments_url()

    def environment_get(self, environment_id):
        return self._build_v1_environments_url() + f"/{environment_id}"


    # # Environment Revision URLs
    def revision_get(self, revision_id):
        return self._build_v4_environments_url() + f"/environmentRevision/{revision_id}"

    def revision_create(self, environment_id):
        return self._build_beta_environments_url() + f"/{environment_id}/revisions"

    def revision_patch(self, environment_id, revision_id):
        return self._build_beta_environments_url() + f"/{environment_id}/revisions/{revision_id}"


    # Deployment URLs
    def deployment_version(self):
        return self.host + "/version"

    # Job URLs
    def job_start(self):
        return f"{self.host}/v4/jobs/start"

    def job_stop(self):
        return f"{self.host}/v4/jobs/stop"

    def jobs_list(self,
                  project_id,
                  order_by,
                  sort_by,
                  page_size,
                  page_no,
                  show_archived,
                  status,
                  tag):

        order_by_query = f"&order_by={order_by}"
        sort_by_query = f"&sort_by={sort_by}"
        page_size_query = f"&page_size={page_size}" if page_size else ""
        page_no_query = f"&page_no={page_no}"
        show_archived_query = "&show_archived=" + show_archived.lower()
        status_from_query = "&status=" + status.lower()
        tag_query = f"&tag={tag}" if tag else ""
        return f"{self.host}/v4/jobs?projectId={project_id}{order_by_query}{sort_by_query}{page_size_query}{page_no_query}{show_archived_query}{status_from_query}{tag_query}"

    def job_status(self, job_id):
        return f"{self.host}/v4/jobs/{job_id}"

    def job_restart(self):
        return f"{self.host}/v4/jobs/restart"

    def job_runtime_execution_details(self, job_id):
        return f"{self.host}/v4/jobs/{job_id}/runtimeExecutionDetails"

    def default_spark_setting(self, project_id):
        return f"{self.host}/v4/jobs/project/{project_id}/defaultSparkSettings"

    def useable_environments_list(self, project_id):
        return f"{self.host}/v4/projects/{project_id}/useableEnvironments"

    # App URLs
    def app_publish(self):
        return self._build_project_url_private_api() + "/nb/startSession"

    # Datasets URLs
    def datasets_list(self, project_id):
        if project_id is None:
            return self.host + "/dataset"
        else:
            return self.host + "/dataset?projectId=" + str(project_id)

    def datasets_create(self):
        return self.host + "/dataset"

    def datasets_details(self, dataset_id):
        return self.host + "/dataset" + "/" + str(dataset_id)

    def datasets_start_upload(self, dataset_id):
        return self.host + f"/v4/datasetrw/datasets/{str(dataset_id)}/snapshot/file/start"

    def datasets_test_chunk(
        self,
        dataset_id,
        upload_key,
        chunk_number,
        total_chunks,
        identifier,
        checksum
    ):
        return (
            self.host +
            f"/v4/datasetrw/datasets/{str(dataset_id)}/snapshot/file/test?key={upload_key}"
            f"&resumableChunkNumber={chunk_number}&resumableIdentifier={quote(identifier)}"
            f"&resumableTotalChunks={total_chunks}&checksum={quote(checksum)}"
        )

    def datasets_upload_chunk(
        self,
        dataset_id,
        key,
        chunk_number,
        total_chunks,
        target_chunk_size,
        current_chunk_size,
        identifier,
        resumable_relative_path,
        checksum
    ):
        return (
            self.host +
            f"/v4/datasetrw/datasets/{dataset_id}/snapshot/file?key={key}&resumableChunkNumber={chunk_number}" +
            f"&resumableChunkSize={target_chunk_size}&resumableCurrentChunkSize={current_chunk_size}"
            f"&resumableIdentifier={quote(identifier)}&resumableRelativePath={quote(resumable_relative_path)}"
            f"&resumableTotalChunks={total_chunks}&checksum={quote(checksum)}"
        )

    def datasets_cancel_upload(self, dataset_id, upload_key):
        return self.host + f"/v4/datasetrw/datasets/{dataset_id}/snapshot/file/cancel/{upload_key}"

    def datasets_end_upload(self, dataset_id, upload_key, target_relative_path=None):
        url = self.host + f"/v4/datasetrw/datasets/{dataset_id}/snapshot/file/end/{upload_key}"
        if target_relative_path:
            url += f"?targetRelativePath={target_relative_path}"
        return url

    def app_list(self, project_id):
        return self.host + f"/v4/modelProducts?projectId={project_id}"

    def app_create(self):
        return self.host + "/v4/modelProducts"

    def app_start(self, app_id):
        return self.host + f"/v4/modelProducts/{app_id}/start"

    def app_stop(self, app_id):
        return self.host + f"/v4/modelProducts/{app_id}/stop"

    def app_get(self, app_id):
        return f"{self.host}/v4/modelProducts/{app_id}"

    # Hardware Tier URLs
    def hardware_tiers_list(self, project_id):
        return self.host + f"/v4/projects/{project_id}/hardwareTiers"

    # Custom Metrics URLs
    def metric_alerts(self):
        return self.host + "/api/metricAlerts/v1"

    def log_metrics(self):
        return self.host + f"/api/metricValues/v1"

    def read_metrics(self, model_monitoring_id, metric):
        return self.host + f"/api/metricValues/v1/{model_monitoring_id}/{metric}"

    # Find Project By OwnerName and project name Url
    def find_project_by_owner_name_and_project_name_url(self):
        return (
            f"{self.host}/v4/gateway/projects/findProjectByOwnerAndName"
            f"?ownerName={self._owner_username}&projectName={self._project_name}"
        )

    # User URLs
    def users_get(self) -> str:
        return self.host + "/v4/users"

    # Budgets and Billing Tags
    def budget_settings(self) -> str:
        return self.host + "/v4/cost/budgets/global/alertsSettings"

    def budgets_default(self, budget_label: Optional[str] = None) -> str:
        label = f"/{budget_label}" if budget_label else ""
        return self.host + f"/v4/cost/budgets/defaults{label}"

    def budget_overrides(self, override_id: Optional[str] = None) -> str:
        override = f"/{override_id}" if override_id else ""
        return self.host + f"/v4/cost/budgets/overrides{override}"

    def billing_tags_settings(self, mode_only: Optional[bool] = False) -> str:
        mode_str = "/mode" if mode_only else ""
        return self.host + "/v4/cost/billingtagSettings" + mode_str

    def billing_tags(self, name: Optional[str] = None) -> str:
        optional_name = f"/{name}" if name else ""
        return self.host + "/v4/cost/billingtags" + optional_name

    def project_billing_tag(self, project_id: Optional[str] = None) -> str:
        return self.host + f"/v4/projects/{project_id}/billingtag"

    def projects_billing_tags(self):
        return self.host + "/v4/projects/billingtags/projects"



import hashlib
import math
import os
import time
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from logging import Logger
from typing import AnyStr

from retry import retry

from domino.http_request_manager import _HttpRequestManager
from domino.routes import _Routes

FILE_UPLOAD_SETTING_DEFAULT = "Ignore"
MAX_WORKERS = 10
MAX_UPLOAD_ATTEMPTS = 10
MB = 2 ** 20  # 2^20 bytes - 1 Megabyte
SLEEP_TIME_IN_SEC = 3
UPLOAD_READ_TIMEOUT_IN_SEC = 30


@dataclass
class UploadChunk:
    """ Class for keeping track of a dataset upload chunk."""
    absolute_path: str
    chunk_number: int
    dataset_id: str
    file_name: str
    file_size: int
    upload_key: str
    identifier: str
    relative_path: str
    target_chunk_size: int
    total_chunks: int


class Uploader:
    def __init__(
        self,
        csrf_no_check_header: {str, str},
        dataset_id: str,
        local_path_to_file_or_directory: str,
        log: Logger,
        request_manager: _HttpRequestManager,
        routes: _Routes,
        target_relative_path: str,

        file_upload_setting: str,
        max_workers: int,
        target_chunk_size: int,
        interrupted: bool = False
    ):
        cleaned_relative_local_path = os.path.relpath(os.path.normpath(local_path_to_file_or_directory), start=os.curdir)
        # in case running on windows
        cleaned_relative_local_path = self._get_unix_style_path(cleaned_relative_local_path)        

        self.csrf_no_check_header = csrf_no_check_header
        self.dataset_id = dataset_id
        self.local_path_file_or_directory = cleaned_relative_local_path
        self.log = log
        self.request_manager = request_manager
        self.routes = routes
        self.target_chunk_size = target_chunk_size or 8 * MB
        self.file_upload_setting = file_upload_setting or FILE_UPLOAD_SETTING_DEFAULT
        self.max_workers = max_workers or MAX_WORKERS
        self.target_relative_path = target_relative_path
        self.interrupted = interrupted
        self.upload_key = None  # this will be set once the session is started

    def __enter__(self):
        # creating upload session
        start_upload_body = {
            "filePaths": [],
            "fileCollisionSetting": self.file_upload_setting
        }
        start_upload_url = self.routes.datasets_start_upload(self.dataset_id)
        self.upload_key = self.request_manager.post(start_upload_url, json=start_upload_body).json()
        if not self.upload_key:
            raise RuntimeError(f"upload key for {self.dataset_id} not found. Session could not start.")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # catching errors
        if exc_type is not None:
            self.log.error(f"Upload for dataset {self.dataset_id} and file or directory "
                           f"`{self.local_path_file_or_directory}` failed, attempting to cancel session. "
                           f"Please try again.")
            self.log.error(f"Error type: {exc_val}. Error message: {exc_tb}.")
            if not isinstance(exc_type, ValueError):
                self._cancel_upload_session()  # is it is a ValueError, canceling session would fail
            return False
        # ending snapshot upload
        try:
            url = self.routes.datasets_end_upload(self.dataset_id, self.upload_key, self.target_relative_path)
            self.request_manager.get(url)
            self.log.info("Upload session ended successfully.")
            return True
        except:
            self.log.error("Ending snapshot upload failed. See error for details. Attempting to cancel "
                           "upload session.")
            self._cancel_upload_session()
            return False

    def upload(self):
        try:
            if not self.upload_key:
                raise RuntimeError(f"upload key for {self.dataset_id} not found. Please start session before uploading.")
            q = self._create_chunk_queue()
            with ThreadPoolExecutor(self.max_workers) as executor:
                # list ensures all the threads are complete before returning results
                results = list(executor.map(self._upload_chunk, q))
            return self.local_path_file_or_directory
        except KeyboardInterrupt:
            self.interrupted = True  # this will allow the threads to stop properly
            raise

    def _cancel_upload_session(self):
        url = self.routes.datasets_cancel_upload(self.dataset_id, self.upload_key)
        self.request_manager.get(url)
        self.log.info("Upload session cancelled successfully.")

    def _create_chunk_queue(self) -> list[UploadChunk]:
        if not os.path.exists(self.local_path_file_or_directory):
            raise ValueError(f"local file or directory {self.local_path_file_or_directory} does not exist.")
        if os.path.isfile(self.local_path_file_or_directory):
            return self._create_chunks(self.local_path_file_or_directory)
        chunk_q = []
        for dirpath, _, filenames in os.walk(self.local_path_file_or_directory):
            for filename in filenames:
                # construct the relative path for each file
                relative_path_to_file = os.path.join(dirpath, filename)

                # in case running on windows
                cleaned_relative_path = self._get_unix_style_path(relative_path_to_file)
                
                # append chunk to queue
                chunk_q.extend(self._create_chunks(cleaned_relative_path))
        return chunk_q

    def _create_chunks(self, local_path_file, starting_index=1) -> list[UploadChunk]:
        file_size = os.path.getsize(local_path_file)
        file_name = os.path.basename(local_path_file)
        total_chunks = max(int(math.ceil(float(file_size) / self.target_chunk_size)), 1)
        return [UploadChunk(absolute_path=os.path.abspath(local_path_file), chunk_number=chunk_num,
                            dataset_id=self.dataset_id, file_name=file_name, file_size=file_size,
                            identifier=f"{file_size}-{file_name}", relative_path=local_path_file,
                            target_chunk_size=self.target_chunk_size, total_chunks=total_chunks,
                            upload_key=self.upload_key)
                for chunk_num in range(starting_index, total_chunks + 1)]

    def _upload_chunk(self, chunk: UploadChunk) -> None:
        if self.interrupted:
            return
        # read the file chunk
        starting_skip = chunk.target_chunk_size * (chunk.chunk_number - 1)
        with open(chunk.absolute_path, 'rb') as file:
            file.seek(starting_skip)
            chunk_data = file.read(chunk.target_chunk_size)

        # testing chunk
        should_upload, checksum = self._test_chunk(chunk, chunk_data)

        # uploading chunk
        if should_upload:
            self._upload_chunk_retry(checksum, chunk, chunk_data)
        else:
            self.log.info(f"Skipping chunk {chunk.chunk_number} of {chunk.total_chunks} for {chunk.file_name}")

    @retry(tries=MAX_UPLOAD_ATTEMPTS, delay=SLEEP_TIME_IN_SEC, backoff=2)
    def _upload_chunk_retry(self, checksum: str, chunk: UploadChunk, chunk_data):
        if self.interrupted:
            return
        actual_chunk_size = len(chunk_data)
        # uploading chunk
        self.log.info(f"Uploading chunk {chunk.chunk_number} of {chunk.total_chunks} for {chunk.file_name}")
        upload_chunk_url = self.routes.datasets_upload_chunk(chunk.dataset_id, chunk.upload_key,
                                                             chunk.chunk_number, chunk.total_chunks,
                                                             chunk.target_chunk_size, actual_chunk_size,
                                                             chunk.identifier, chunk.relative_path,
                                                             checksum)
        # files to pass in post's **kwargs
        files = {
            chunk.relative_path: (chunk.file_name, chunk_data, 'application/octet-stream')
        }
        start_time_ns = time.time_ns()
        # making call to upload
        self.request_manager.post(upload_chunk_url, files=files, timeout=UPLOAD_READ_TIMEOUT_IN_SEC,
                                  headers=self.csrf_no_check_header)
        end_time_ns = time.time_ns()
        duration_ns = end_time_ns - start_time_ns
        bandwidth_bytes_per_second = actual_chunk_size / duration_ns * 1000000000.0
        self.log.info(f"Uploaded chunk {chunk.chunk_number} of {chunk.total_chunks} for {chunk.file_name} "
                      f"in {duration_ns / 1_000_000:.1f}ms ({bandwidth_bytes_per_second:.1f} B/s)")

    def _test_chunk(self, chunk: UploadChunk, chunk_data: AnyStr) -> (bool, int):
        # computing the MD5 checksum
        digest = hashlib.md5()
        digest.update(chunk_data)

        chunk_checksum = digest.hexdigest().upper()
        # testing chunk
        test_chunk_url = self.routes.datasets_test_chunk(chunk.dataset_id, chunk.upload_key, chunk.chunk_number,
                                                         chunk.total_chunks, chunk.identifier, chunk_checksum)
        # test chunk returns no content if it should upload
        return self.request_manager.get(test_chunk_url).status_code == 204, chunk_checksum

    def _get_unix_style_path(self, path: str) -> str:
        # when running on Windows, converts path to Unix-style path, which the upload chunk API expects
        if os.sep != '/':
            path = path.replace(os.sep, '/')
        return path



"""
To run these tests you need to have at least configured airflow in
local mode and run:

`airflow db init`
"""
import os
from datetime import datetime
from airflow import DAG
from airflow.models import TaskInstance
import pytest
from pprint import pformat

from domino.airflow import DominoSparkOperator
from domino.exceptions import RunFailedException
from domino.helpers import domino_is_reachable

TEST_PROJECT = os.environ.get("DOMINO_SPARK_TEST_PROJECT")
dag_id = "test_spark_operator"

@pytest.mark.skipif(os.getenv("SPARK_DEP") != "yes", reason="Extra dependency required")
@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_spark_operator_no_cluster():
    execution_dt = datetime.now()

    dag = DAG(dag_id, start_date=execution_dt)
    task = DominoSparkOperator(
        dag=dag,
        task_id="test_spark_operator_no_cluster",
        project=TEST_PROJECT,
        command="test_spark.py",
    )
    task.run()
    ti = TaskInstance(task=task, execution_date=execution_dt)
    task.execute(ti.get_template_context())

@pytest.mark.skipif(os.getenv("SPARK_DEP") != "yes", reason="Extra dependency required")
@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_spark_operator_with_cluster(spark_cluster_env_id):

    execution_dt = datetime.now()
    dag = DAG(dag_id, start_date=execution_dt)
    task = DominoSparkOperator(
        dag=dag,
        task_id="test_spark_operator_with_cluster",
        project=TEST_PROJECT,
        command="test_spark.py",
        on_demand_spark_cluster_properties={
            "computeEnvironmentId": spark_cluster_env_id,
            "executorCount": 3,
        },
    )
    task.run()
    ti = TaskInstance(task=task, execution_date=execution_dt)
    task.execute(ti.get_template_context())

@pytest.mark.skipif(os.getenv("SPARK_DEP") != "yes", reason="Extra dependency required")
@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_spark_operator_with_compute_cluster_properties(spark_cluster_env_id):
    execution_dt = datetime.now()

    dag = DAG(dag_id, start_date=execution_dt)
    task = DominoSparkOperator(
        dag=dag,
        task_id="test_spark_operator_with_compute_cluster_properties",
        project=TEST_PROJECT,
        command="test_spark.py",
        compute_cluster_properties={
            "clusterType": "Spark",
            "computeEnvironmentId": spark_cluster_env_id,
            "masterHardwareTierId": "small-k8s",
            "workerHardwareTierId": "small-k8s",
            "workerCount": 3,
        },
    )
    task.run()
    ti = TaskInstance(task=task, execution_date=execution_dt)
    task.execute(ti.get_template_context())

@pytest.mark.skipif(os.getenv("SPARK_DEP") != "yes", reason="Extra dependency required")
@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_spark_operator_no_cluster_failed():
    execution_dt = datetime.now()

    dag = DAG(dag_id, start_date=execution_dt)
    task = DominoSparkOperator(
        dag=dag,
        task_id="test_spark_operator_no_cluster_failed",
        project=TEST_PROJECT,
        command="test_spark_fail.sh",
    )

    with pytest.raises(RunFailedException):
        task.run()
        ti = TaskInstance(task=task, execution_date=execution_dt)
        task.execute(ti.get_template_context())



from abc import ABC, abstractmethod
from copy import deepcopy
import json
from typing import Dict, List, Any, Optional, Union

from ._impl.custommetrics.model.metric_value_v1 import MetricValueV1
from ._impl.custommetrics.model.metric_alert_request_v1 import MetricAlertRequestV1
from ._impl.custommetrics.model.target_range_v1 import TargetRangeV1
from ._impl.custommetrics.paths.api_metric_alerts_v1.post import (
    request_body_metric_alert_request_v1,
)
from ._impl.custommetrics.model.new_metric_values_envelope_v1 import (
    NewMetricValuesEnvelopeV1,
)
from ._impl.custommetrics.model.new_metric_value_v1 import NewMetricValueV1
from ._impl.custommetrics.model.metric_tag_v1 import MetricTagV1
from ._impl.custommetrics.paths.api_metric_values_v1.post import (
    request_body_new_metric_values_envelope_v1,
)
from ._impl.custommetrics.paths.api_metric_values_v1_model_monitoring_id_metric.get import (
    _response_for_200,
    ApiResponseFor200,
)
from ._impl.custommetrics.model.metric_values_envelope_v1 import MetricValuesEnvelopeV1
from ._impl.custommetrics.model.metadata_v1 import MetadataV1
from ._impl.custommetrics.api_client import SerializedRequestBody
from ._impl.custommetrics import schemas
from ._impl.custommetrics import configuration


class _CustomMetricsClientBase(ABC):
    # Target range values
    LESS_THAN = "lessThan"
    LESS_THAN_EQUAL = "lessThanEqual"
    GREATER_THAN = "greaterThan"
    GREATER_THAN_EQUAL = "greaterThanEqual"
    BETWEEN = "between"

    @abstractmethod
    def trigger_alert(
        self,
        model_monitoring_id: str,
        metric: str,
        value: Any,
        condition: str = None,
        lower_limit: Any = None,
        upper_limit: Any = None,
        description: str = None,
    ) -> None:
        pass

    @abstractmethod
    def log_metric(
        self,
        model_monitoring_id: str,
        metric: str,
        value: Any,
        timestamp: str,
        tags: Dict = None,
    ) -> None:
        pass

    @abstractmethod
    def log_metrics(self, metric_values_array: List) -> None:
        pass

    @abstractmethod
    def read_metrics(
        self,
        model_monitoring_id: str,
        metric: str,
        start_timestamp: str,
        end_timestamp: str,
    ) -> Dict:
        pass


class _CustomMetricsClientGen(_CustomMetricsClientBase):
    """
    Client that partially re-uses the OpenAPI auto-generated code.
    """

    def __init__(self, parent, routes):
        self._parent = parent
        self._routes = routes

    def trigger_alert(
        self,
        model_monitoring_id: str,
        metric: str,
        value: Any,
        condition: str = None,
        lower_limit: Any = None,
        upper_limit: Any = None,
        description: str = None,
    ) -> None:
        url = self._routes.metric_alerts()
        target_range: Optional[TargetRangeV1] = (
            None
            if condition is None
            else TargetRangeV1(
                condition=condition, lowerLimit=lower_limit, upperLimit=upper_limit
            )
        )
        req = MetricAlertRequestV1(
            modelMonitoringId=model_monitoring_id,
            metric=metric,
            value=value,
            targetRange=target_range, # type: ignore
            description=description,
        )
        ser_body: SerializedRequestBody = (
            request_body_metric_alert_request_v1.serialize(req, "application/json")
        )
        json_data = json.loads(
            ser_body["body"].decode("utf-8") # type: ignore
        )  # extra work to reuse request_manager
        self._parent.request_manager.post(url, json=json_data)

    def log_metric(
        self,
        model_monitoring_id: str,
        metric: str,
        value: Any,
        timestamp: str,
        tags: Dict = None,
    ) -> None:
        item = {
            "modelMonitoringId": model_monitoring_id,
            "metric": metric,
            "value": value,
            "timestamp": timestamp,
        }
        if tags is not None:
            item["tags"] = tags
        self.log_metrics([item])

    def _to_new_metric_value(self, item: Dict) -> NewMetricValueV1:
        tags: Union[List[MetricTagV1],schemas.Unset] = schemas.unset
        if "tags" in item:
            tags = [MetricTagV1(key=k, value=v) for k, v in item["tags"].items()]
        ret = NewMetricValueV1(
            referenceTimestamp=item["timestamp"],
            metric=item["metric"],
            modelMonitoringId=item["modelMonitoringId"],
            value=item["value"],
            tags=tags,
        )
        return ret

    def log_metrics(self, metric_values_array: List) -> None:
        url = self._routes.log_metrics()
        newMetricValues = [self._to_new_metric_value(x) for x in metric_values_array]
        req = NewMetricValuesEnvelopeV1(newMetricValues=newMetricValues)
        ser_body: SerializedRequestBody = (
            request_body_new_metric_values_envelope_v1.serialize(
                req, "application/json"
            )
        )
        json_data = json.loads(
            ser_body["body"].decode("utf-8") # type: ignore
        )  # extra work to reuse request_manager
        self._parent.request_manager.post(url, json=json_data)

    def _from_metric_value(self, obj: MetricValueV1) -> Dict:
        # Numbers come in as Decimal objects but we want floats
        ret = {"timestamp": str(obj.referenceTimestamp), "value": float(obj["value"])}
        if obj.tags is not None:
            ret["tags"] = {tag["key"]: tag["value"] for tag in obj.tags}
        return ret

    def read_metrics(
        self,
        model_monitoring_id: str,
        metric: str,
        start_timestamp: str,
        end_timestamp: str,
    ) -> Dict:
        url = self._routes.read_metrics(model_monitoring_id, metric)
        params = {
            "startingReferenceTimestampInclusive": start_timestamp,
            "endingReferenceTimestampInclusive": end_timestamp,
        }
        res = self._parent.request_manager.get(url, params=params).json()
        mvs: MetricValuesEnvelopeV1 = MetricValuesEnvelopeV1.from_openapi_data_oapg(res)
        ret = {
            "metadata": dict(mvs.metadata),
            "metricValues": [self._from_metric_value(x) for x in mvs.metricValues],
        }
        return ret


class _CustomMetricsClient(_CustomMetricsClientBase):
    """
    Hand-rolled client kept in case of unexpected problems.
    """

    def __init__(self, parent, routes):
        self._parent = parent
        self._routes = routes

    def trigger_alert(
        self,
        model_monitoring_id: str,
        metric: str,
        value: Any,
        condition: str = None,
        lower_limit: Any = None,
        upper_limit: Any = None,
        description: str = None,
    ) -> None:
        url = self._routes.metric_alerts()
        request = {
            "modelMonitoringId": model_monitoring_id,
            "metric": metric,
            "value": value,
            "targetRange": {"condition": condition},
        }
        if condition:
            request["targetRange"] = {"condition": condition}
            if lower_limit:
                request["targetRange"]["lowerLimit"] = lower_limit
            if upper_limit:
                request["targetRange"]["upperLimit"] = upper_limit
        if description:
            request["description"] = description
        self._parent.request_manager.post(url, json=request)

    def log_metric(
        self,
        model_monitoring_id: str,
        metric: str,
        value: Any,
        timestamp: str,
        tags: Dict = None,
    ) -> None:
        item = {
            "modelMonitoringId": model_monitoring_id,
            "metric": metric,
            "value": value,
            "timestamp": timestamp,
        }
        if tags is not None:
            item["tags"] = tags
        self.log_metrics([item])

    def _rewrite_tags_in(self, item) -> List:
        return [{"key": k, "value": v} for k, v in item["tags"].items()]

    def _rewrite_tags_out(self, item) -> Dict:
        return {v["key"]: v["value"] for v in item["tags"]}

    def log_metrics(self, metric_values_array: List) -> None:
        url = self._routes.log_metrics()
        new_values = []
        for item in metric_values_array:
            new_item = deepcopy(item)
            if "tags" in new_item:
                new_item["tags"] = self._rewrite_tags_in(new_item)
            if "timestamp" in new_item:
                new_item["referenceTimestamp"] = new_item["timestamp"]
                del new_item["timestamp"]
            new_values.append(new_item)

        request = {"newMetricValues": new_values}
        self._parent.request_manager.post(url, json=request)

    def read_metrics(
        self,
        model_monitoring_id: str,
        metric: str,
        start_timestamp: str,
        end_timestamp: str,
    ) -> Dict:
        url = self._routes.read_metrics(model_monitoring_id, metric)
        params = {
            "startingReferenceTimestampInclusive": start_timestamp,
            "endingReferenceTimestampInclusive": end_timestamp,
        }
        res = self._parent.request_manager.get(url, params=params).json()
        if "metricValues" in res:
            for item in res["metricValues"]:
                if "referenceTimestamp" in item:
                    item["timestamp"] = item["referenceTimestamp"]
                    del item["referenceTimestamp"]
                if "tags" in item:
                    item["tags"] = self._rewrite_tags_out(item)
        return res



import os


def get_is_production() -> bool:
    return os.environ.get("DOMINO_AI_SYSTEM_IS_PROD", "false").lower() == "true"


def _get_ai_system_id() -> str | None:
    return os.environ.get("DOMINO_APP_ID")


def is_ai_system() -> bool:
    return get_is_production() and _get_ai_system_id() is not None


def build_ai_system_experiment_name(id: str) -> str:
    return f"ai_system_experiment_{id}"


def get_running_ai_system_experiment_name() -> str | None:
    if is_ai_system():
        return build_ai_system_experiment_name(_get_ai_system_id())
    return None



import logging as logger
import os
from docker.types import Mount
import polling2
import pytest
import shutil
from unittest.mock import patch

from ...conftest import TEST_AI_SYSTEMS_ENV_VARS
from domino.aisystems._constants import MIN_MLFLOW_VERSION
from .test_util import reset_prod_tracing

@pytest.fixture
def tracing():
        pytest.importorskip("mlflow")
        import domino.aisystems.tracing as tracing
        yield tracing

@pytest.fixture
def logging():
        pytest.importorskip("mlflow")
        import domino.aisystems.logging as logging
        yield logging

@pytest.fixture
def mlflow():
        pytest.importorskip("mlflow")
        import mlflow
        yield mlflow

def _remove_mlruns_data():
        try:
                shutil.rmtree('./mlruns')
        except Exception as e:
                logger.warning(f"Failed to remove mlfruns directory during test cleanup: {e}")

@pytest.fixture(scope="package")
def setup_mlflow_tracking_server_no_env_var_mock(docker_client):
        pytest.importorskip("mlflow")
        from mlflow import MlflowClient

        with patch("domino.aisystems._verify_domino_support.verify_domino_support", clear=True) as mock_verify_domino_support:
                mock_verify_domino_support.return_value = None
                container_name = "test_mlflow_tracking_server"
                docker_client.containers.run(
                        f"ghcr.io/mlflow/mlflow:v{MIN_MLFLOW_VERSION}",
                        detach=True,
                        name=container_name,
                        ports={5000:5000},
                        command="mlflow ui --host 0.0.0.0 --serve-artifacts",
                )

                try:
                        live_container = polling2.poll(
                                target=lambda: docker_client.containers.get(container_name),
                                check_success=lambda container: container.status == 'running',
                                timeout=10,
                                step=2,
                                ignore_exceptions=True,
                        )

                        # verify api is reachable
                        client = MlflowClient()
                        experiments = polling2.poll(
                                target=lambda: client.search_experiments(),
                                check_success=lambda exp: True,
                                timeout=10,
                                step=2,
                                ignore_exceptions=True,
                        )

                        yield live_container
                        live_container.remove(force=True)
                        _remove_mlruns_data()
                except Exception as e:
                        live_container = docker_client.containers.get(container_name)
                        container_status = live_container.status
                        logger.error(f'Mlflow tracking server did not get to running state. status: {container_status}')
                        logger.info(live_container.logs())
                        live_container.remove(force=True)
                        _remove_mlruns_data()
                        raise e

@pytest.fixture
def setup_mlflow_tracking_server(setup_mlflow_tracking_server_no_env_var_mock, docker_client):
        with patch.dict(os.environ, TEST_AI_SYSTEMS_ENV_VARS, clear=True):
                yield setup_mlflow_tracking_server_no_env_var_mock



import json

from .._client import client
from .._eval_tags import build_eval_result_tag, validate_label
from .._verify_domino_support import verify_domino_support
from .._constants import DOMINO_INTERNAL_EVAL_TAG


def add_domino_tags(trace_id: str):
    """
    Tags a trace as one that contains an evaluation
    Args:
        trace_id: string, the ID of the trace to tag
    """
    client.set_trace_tag(trace_id, DOMINO_INTERNAL_EVAL_TAG, json.dumps(True))


def log_evaluation(
        trace_id: str,
        name: str,
        value: float | str,
    ):
    """This logs evaluation data and metadata to a parent trace. This is used to log the evaluation of a span
    after it was created. This is useful for analyzing past performance of an AI System component.

    Args:
        trace_id: the ID of the trace to evaluate

        name: a label for the evaluation result. This is used to identify the evaluation result

        value: the evaluation result to log. This must be a float or string
    """
    verify_domino_support()
    validate_label(name)

    if value is not None:
        formatted_value = value
        if not isinstance(value, (str)):
            formatted_value = json.dumps(value)

        client.set_trace_tag(
            trace_id,
            build_eval_result_tag(name, value),
            formatted_value,
        )
        add_domino_tags(trace_id)


# TODO log feedback



import os

import pytest
import requests

import domino.authentication
from domino import Domino, authentication
from domino.constants import (
    DOMINO_HOST_KEY_NAME,
    DOMINO_TOKEN_FILE_KEY_NAME,
    DOMINO_USER_API_KEY_KEY_NAME,
)
from domino.helpers import domino_is_reachable


@pytest.mark.usefixtures("clear_api_key_from_env", "clear_token_file_from_env")
def test_no_auth_type_error():
    """
    Confirm that trying to instantiate a Domino object with no auth raises
    an AssertionError with the expected error message.
    """
    dummy_host = "http://domino.somefakecompany.com"

    with pytest.raises(RuntimeError):
        Domino(host=dummy_host, project="anyuser/quick-start")


def test_invalid_token_file_error():
    """
    Confirm that trying to instantiate a Domino object with an invalid token file
    raises a FileNotFoundError exception.
    """
    dummy_host = "http://domino.somefakecompany.com"
    invalid_file = "non_existent_token_file.txt"

    with pytest.raises(FileNotFoundError):
        test_domino = Domino(
            host=dummy_host,
            project="anyuser/quick-start",
            domino_token_file=invalid_file,
        )
        test_domino.runs_list()["data"]


def test_malformed_project_input_error(caplog):
    """
    Confirm that passing in a project name without the user name raises a ValueError.
    """
    # This test uses the provided pytest fixture, caplog. For more info, see:
    # https://docs.pytest.org/en/6.2.x/reference.html?highlight=caplog#std-fixture-caplog
    dummy_host = "http://domino.somefakecompany.com"
    invalid_project = "user_project_name"

    with pytest.raises(ValueError):
        Domino(host=dummy_host, project=invalid_project)

    err_msg = f"{invalid_project} must be given in the form username/projectname"
    assert any(
        err_msg in rec.message for rec in caplog.records
    ), f"Expected: {err_msg}, got {caplog.records}"


@pytest.mark.usefixtures("mock_domino_version_response")
def test_object_creation_with_auth_token():
    """
    Confirm that the expected auth type is used when using an auth token.
    """
    dummy_host = "http://domino.somefakecompany.com"
    dummy_auth_token = "top_secret_auth_token"

    d = Domino(
        host=dummy_host, project="anyuser/quick-start", auth_token=dummy_auth_token
    )
    assert isinstance(
        d.request_manager.auth, authentication.BearerAuth
    ), "Authentication using auth_token should be of type authentication.BearerAuth"


@pytest.mark.usefixtures("mock_domino_version_response", "clear_token_file_from_env")
def test_object_creation_with_api_key():
    """
    Confirm that the expected auth type is used when using an api key.
    """
    dummy_host = "http://domino.somefakecompany.com"
    dummy_api_key = "top_secret_api_key"

    d = Domino(host=dummy_host, project="anyuser/quick-start", api_key=dummy_api_key)
    assert isinstance(
        d.request_manager.auth, requests.auth.HTTPBasicAuth
    ), "Authentication using API key should be of type requests.auth.HTTPBasicAuth"


@pytest.mark.usefixtures("mock_domino_version_response", "clear_token_file_from_env", "mock_proxy_response")
def test_object_creation_with_api_proxy():
    """
    Confirm that the expected auth type is used when using api proxy.
    """
    dummy_host = "http://domino.somefakecompany.com"
    dummy_api_proxy = "localhost:1234"

    d = Domino(host=dummy_host, project="anyuser/quick-start", api_proxy=dummy_api_proxy)
    assert isinstance(
        d.request_manager.auth, domino.authentication.ProxyAuth
    ), "Authentication using API proxy should be of type domino.authentication.ProxyAuth"
    assert d.request_manager.auth.api_proxy == "http://localhost:1234"

@pytest.mark.usefixtures("mock_domino_version_response", "clear_token_file_from_env", "mock_proxy_response_https")
def test_object_creation_with_api_proxy_with_scheme():
    """
    Confirm that the expected auth type is used when using api proxy.
    """
    dummy_host = "http://domino.somefakecompany.com"
    dummy_api_proxy = "https://localhost:1234"

    d = Domino(host=dummy_host, project="anyuser/quick-start", api_proxy=dummy_api_proxy)
    assert isinstance(
        d.request_manager.auth, domino.authentication.ProxyAuth
    ), "Authentication using API proxy should be of type domino.authentication.ProxyAuth"
    assert d.request_manager.auth.api_proxy == "https://localhost:1234"


@pytest.mark.usefixtures("mock_domino_version_response")
def test_object_creation_with_token_file(dummy_token_file):
    """
    Confirm that the expected auth type is used when using a token file.
    """
    dummy_host = "http://domino.somefakecompany.com"

    d = Domino(
        host=dummy_host,
        project="anyuser/quick-start",
        domino_token_file=dummy_token_file,
    )
    assert isinstance(
        d.request_manager.auth, authentication.BearerAuth
    ), "Authentication using token_file should be of type authentication.BearerAuth"


@pytest.mark.usefixtures("mock_domino_version_response", "clear_token_file_from_env")
def test_re_authentication_with_a_new_type(dummy_token_file):
    """
    Confirm that the client can re-authenticate with a new authentication token,
    or (in this case) even switch to an API key.
    """
    dummy_host = "http://domino.somefakecompany.com"

    d = Domino(
        host=dummy_host,
        project="anyuser/quick-start",
        domino_token_file=dummy_token_file,
    )
    assert isinstance(
        d.request_manager.auth, authentication.BearerAuth
    ), "Authentication using token_file should be of type authentication.BearerAuth"

    d.authenticate(api_key="bogus_api_key")
    assert isinstance(
        d.request_manager.auth, requests.auth.HTTPBasicAuth
    ), "Re-authentication with API key should be of type requests.auth.HTTPBasicAuth"


@pytest.mark.usefixtures("mock_domino_version_response")
def test_auth_token_precedence():
    """
    Confirm that auth token takes precedence over both token file and API key authentication.
    """
    dummy_host = "http://domino.somefakecompany.com"
    dummy_auth_token = "top_secret_auth_token"
    invalid_token_file = "non_existent_token_file.txt"
    dummy_api_key = "top_secret_api_key"

    d = Domino(
        host=dummy_host,
        project="anyuser/quick-start",
        auth_token=dummy_auth_token,
        domino_token_file=invalid_token_file,
        api_key=dummy_api_key,
    )
    assert isinstance(
        d.request_manager.auth, authentication.BearerAuth
    ), "With multiple authentication options, auth_token string should take highest precendence"


@pytest.mark.usefixtures("mock_domino_version_response")
@pytest.mark.skipif(
    not os.getenv(DOMINO_TOKEN_FILE_KEY_NAME), reason="No token file in environment"
)  # noqa:E501
def test_auth_with_api_key_and_env_token_file():
    """
    Confirm that api key takes precedence over both token file and API key authentication
    if they are not passed as parameters.
    """
    dummy_host = "http://domino.somefakecompany.com"
    dummy_api_key = "top_secret_api_key"

    d = Domino(host=dummy_host, project="anyuser/quick-start", api_key=dummy_api_key)
    assert isinstance(
        d.request_manager.auth, requests.auth.HTTPBasicAuth
    ), "With only api key passed as a parameter, it takes precedence over DOMINO_TOKEN_FILE_KEY_NAME"


@pytest.mark.usefixtures("mock_domino_version_response")
def test_token_file_precedence(dummy_token_file):
    """
    Confirm that token file authentication takes precedence over API key authentication.
    """
    dummy_host = "http://domino.somefakecompany.com"
    dummy_api_key = "top_secret_api_key"

    d = Domino(
        host=dummy_host,
        project="anyuser/quick-start",
        domino_token_file=dummy_token_file,
        api_key=dummy_api_key,
    )
    assert isinstance(
        d.request_manager.auth, authentication.BearerAuth
    ), "When a token file and API are present, auth by token file should take highest precendence"


@pytest.mark.usefixtures("clear_token_file_from_env")
@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
@pytest.mark.skipif(
    not os.getenv(DOMINO_USER_API_KEY_KEY_NAME), reason="No API key in environment"
)  # noqa:E501
def test_auth_against_real_deployment_with_api_key():
    """
    Confirm against a live system that validating by API key works.

    Assumes that ${DOMINO_API_HOST} contains a valid Domino URL
    Assumes that ${DOMINO_USER_API_KEY} contains a valid API key
    """
    host = os.getenv(DOMINO_HOST_KEY_NAME)
    api_key = os.getenv(DOMINO_USER_API_KEY_KEY_NAME)

    d = Domino(host=host, project="anyuser/quick-start", api_key=api_key)
    assert isinstance(
        d.request_manager.auth, requests.auth.HTTPBasicAuth
    ), "Authentication using API key should be of type requests.auth.HTTPBasicAuth"

    # Raises a requests.exceptions.HTTPError if authentication failed
    d.environments_list()


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
@pytest.mark.skipif(
    not os.getenv(DOMINO_TOKEN_FILE_KEY_NAME), reason="No token file in environment"
)  # noqa:E501
def test_auth_against_real_deployment_with_token_file():
    """
    Confirm against a live system that validating by token file works.

    Assumes that ${DOMINO_API_HOST} contains a valid Domino URL
    Assumes that ${DOMINO_TOKEN_FILE} is a path to a file that contains a valid token
    """
    host = os.getenv(DOMINO_HOST_KEY_NAME)
    token_file = os.getenv(DOMINO_TOKEN_FILE_KEY_NAME)

    d = Domino(host=host, project="anyuser/quick-start", domino_token_file=token_file)
    assert isinstance(
        d.request_manager.auth, authentication.BearerAuth
    ), "Authentication using token_file should be of type authentication.BearerAuth"

    # Raises a requests.exceptions.HTTPError if authentication failed
    d.environments_list()



from .tracing import (
    add_tracing,
    SpanSummary,
    TraceSummary,
    SearchTracesResponse,
    search_traces,
    search_ai_system_traces,
)
from .inittracing import init_tracing as init_tracing

__all__ = [
    "add_tracing",
    "SpanSummary",
    "TraceSummary",
    "SearchTracesResponse",
    "search_traces",
    "search_ai_system_traces",
    "init_tracing",
]



import pytest

from domino.aisystems._eval_tags import InvalidEvaluationLabelException
from .mlflow_fixtures import fixture_create_traces

def test_log_evaluation_dev(setup_mlflow_tracking_server, mlflow, logging):
	# create experiment
	exp = mlflow.set_experiment("test_log_evaluation")

	fixture_create_traces()

	# log evaluations to traces
	traces = mlflow.search_traces(experiment_ids=[exp.experiment_id], filter_string="trace.name = 'test_add'", return_type='list')
	for trace in traces:
		logging.log_evaluation(
			trace.info.trace_id,
			value=1,
			name="helpfulness",
		)
		logging.log_evaluation(
			trace.info.trace_id,
			value="dogs",
			name="category",
		)

	# verify tags on traces
	tagged_traces = mlflow.search_traces(experiment_ids=[exp.experiment_id], filter_string="trace.name = 'test_add'", return_type="list")
	tags = tagged_traces[0].info.tags

	assert tags['domino.prog.label.category'] == 'dogs'
	assert tags['domino.prog.metric.helpfulness'] == '1'
	assert tags['domino.internal.is_eval'] == 'true'

def test_log_evaluation_invalid_name(setup_mlflow_tracking_server, mlflow, logging):
	# create experiment
	exp = mlflow.set_experiment("test_log_evaluation_invalid_name")

	fixture_create_traces()

	# log evaluations to traces
	traces = mlflow.search_traces(experiment_ids=[exp.experiment_id], filter_string="trace.name = 'test_add'", return_type='list')
	trace = traces[0]

	with pytest.raises(InvalidEvaluationLabelException):
		logging.log_evaluation(
			trace.info.trace_id,
			value=1,
			name="*",
		)

def test_log_evaluation_non_string_float(setup_mlflow_tracking_server, mlflow, logging):
	"""
	Log evaluation should not allow logging objects
	"""
	# create experiment
	exp = mlflow.set_experiment("test_log_evaluation_non_string_float")

	fixture_create_traces()

	# log evaluations to traces
	traces = mlflow.search_traces(experiment_ids=[exp.experiment_id], filter_string="trace.name = 'test_add'", return_type='list')
	trace = traces[0]

	with pytest.raises(TypeError):
		logging.log_evaluation(
			trace.info.trace_id,
			value={},
			name="myobject",
		)
